"""
MCP 服务器

查询工具（3 个）：
0. get_devices - 列出所有设备
1. get_status - 设备全量状态（合并 10 个 trait 查询）
2. get_map_content - 地图内容（独立重量级请求）

控制工具（10 个）：
3. start_clean - 开始清洁（全屋或指定房间）
4. stop_clean - 停止清洁
5. pause_clean - 暂停清洁
6. resume_clean - 恢复清洁
7. go_home - 回充电座
8. find_robot - 发出声音定位机器人
9. dock_action - 基座操作（集尘/洗拖布）
10. set_volume - 调节音量
11. get_routines - 列出智能场景
12. execute_routine - 执行智能场景
"""

import asyncio
import functools
import logging
from enum import Enum
from typing import Any

from mcp.server.fastmcp import FastMCP
from roborock import RoborockCommand
from roborock.exceptions import RoborockException

from .connection import (
    AmbiguousDeviceError,
    AuthExpiredError,
    ConnectionManager,
    DeviceNotFoundError,
    NoDeviceError,
    lifespan,
)

logger = logging.getLogger(__name__)


def _handle_errors(func):
    """包装工具函数，将已知异常转为结构化错误响应"""
    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except (AuthExpiredError, NoDeviceError, AmbiguousDeviceError,
                DeviceNotFoundError, ValueError) as e:
            return {"error": str(e)}
        except RoborockException as e:
            logger.warning("工具 %s 设备通信失败: %s", func.__name__, e)
            return {"error": f"设备通信失败: {type(e).__name__}: {e}"}
        except Exception as e:
            logger.exception("工具 %s 执行失败", func.__name__)
            return {"error": f"内部错误: {type(e).__name__}: {e}"}
    return wrapper

# 创建 FastMCP 实例
mcp = FastMCP(
    "roborock-mcp",
    lifespan=lifespan,
)


def get_conn() -> ConnectionManager:
    """从 MCP context 获取 ConnectionManager"""
    return mcp.get_context().request_context.lifespan_context["conn"]


async def _safe_refresh(trait, name: str) -> bool:
    """安全刷新单个 trait，失败不影响其他"""
    try:
        await trait.refresh()
        return True
    except Exception as e:
        logger.warning("Trait %s refresh 失败: %s", name, e)
        return False


def _serialize_value(value: Any) -> Any:
    """将 Enum 等特殊类型转为可序列化值"""
    if value is None:
        return None
    if isinstance(value, Enum):
        return value.name
    if isinstance(value, list):
        return [_serialize_value(v) for v in value]
    return value


def _field(value: Any, description: str) -> dict[str, Any]:
    """包装字段值和描述，供 LLM 理解"""
    return {"value": value, "description": description}


# StatusV2 白名单：仅输出有意义的字段
# state/error_code 改在 computed properties 中输出（避免与 state_name/error_code_name 重复）
_STATUS_FIELDS: dict[str, str] = {
    # 核心状态
    "battery": "电池电量 (%)",
    "charge_status": "充电状态 (0=未充电, 1=充电中, 2=充电完成)",
    # 清洁进度
    "clean_percent": "清洁任务完成百分比 (%)",
    "in_cleaning": "清洁状态 (complete/global_clean_not_complete/zone_clean_not_complete/segment_clean_not_complete)",
    "in_returning": "是否在返回途中 (0=否, 1=是)",
    "avoid_count": "当前任务避障次数",
    "repeat": "当前任务重复遍数",
    # 地图
    "map_present": "是否有地图 (0=无, 1=有)",
    # 基座
    "dock_type": "基座类型 (no_dock/auto_empty_dock/empty_wash_fill_dock/p10_pro_dock 等)",
    "dock_error_status": "基座错误状态 (ok/duct_blockage/water_empty/waste_water_tank_full 等)",
    "dust_collection_status": "集尘状态 (0=未集尘, 1=集尘中)",
    "auto_dust_collection": "自动集尘开关 (0=关闭, 1=开启)",
    # 洗拖布/烘干
    "wash_status": "洗拖布状态 (0=空闲, 1=洗涤中)",
    "wash_phase": "洗拖布阶段 (0=空闲)",
    "wash_ready": "是否准备好洗拖布 (0=否, 1=是)",
    "dry_status": "热风烘干状态 (0=关闭, 1=烘干中)",
    # 水箱
    "water_box_status": "水箱安装状态 (0=未安装, 1=已安装)",
    "water_box_carriage_status": "水箱托架状态 (0=正常)",
    "water_shortage_status": "缺水状态 (0=正常)",
    # 其他
    "is_locating": "是否正在定位 (0=否, 1=是)",
    "lock_status": "童锁状态 (0=未锁定, 1=锁定)",
    "mop_forbidden_enable": "拖地禁区是否启用 (0=关闭, 1=启用)",
    "collision_avoid_status": "避障功能 (0=关闭, 1=开启)",
}


# ============================================================================
# 工具定义
# ============================================================================

@mcp.tool(
    annotations={
        "title": "List Roborock Devices",
        "readOnlyHint": True,
        "destructiveHint": False,
        "idempotentHint": True,
        "openWorldHint": True,
    }
)
@_handle_errors
async def get_devices() -> dict[str, Any]:
    """列出账号下所有 Roborock 设备。

    通过 Roborock 云端 API 获取设备列表（不走 MQTT），返回每台设备的基本信息。

    Returns:
        dict: {
            "count": int,           # 设备数量
            "devices": [
                {
                    "name": str,    # 设备名称（用户在 App 中设置的名称）
                    "duid": str,    # 设备唯一标识符
                    "model": str,   # 设备型号（如 "roborock.vacuum.a75"）
                    "firmware": str # 固件版本号
                }
            ]
        }
    """
    conn = get_conn()
    devices = await conn.list_devices()

    result = {
        "count": len(devices),
        "devices": []
    }

    for device in devices:
        info = {
            "name": device.name,
            "duid": device.duid,
            "model": device.product.model if device.product else None,
            "firmware": device.device_info.fv if device.device_info else None,
        }
        result["devices"].append(info)

    return result


@mcp.tool(
    annotations={
        "title": "Get Device Full Status",
        "readOnlyHint": True,
        "destructiveHint": False,
        "idempotentHint": True,
        "openWorldHint": True,
    }
)
@_handle_errors
async def get_status(device_name: str | None = None) -> dict[str, Any]:
    """获取 Roborock 设备精简状态（通过 MQTT 实时查询 8 个 trait）。

    每个字段返回 {"value": x, "description": "..."} 结构，方便理解字段含义。
    并行查询设备信息，包含：
    - 实时状态：电量、吸力、清洁进度、基座状态、错误码等（已精简，去除内部字段和重复字段）
    - 勿扰模式：开关状态及时间段
    - 清洁历史：累计清洁时间/面积/次数，最近一次清洁详情
    - 音量、房间列表（含 segment_id）、地图列表、耗材使用时间

    单个 trait 查询失败不影响其他 trait 返回，失败的 trait 对应字段会缺失。

    Args:
        device_name: 设备名称，支持模糊匹配。单设备时可省略，自动选择唯一设备。
    """
    conn = get_conn()
    device = await conn.get_device(device_name)

    if not device.v1_properties:
        return {"error": "设备不支持 V1 协议"}

    props = device.v1_properties

    # 并行刷新所有 trait（device_features 保留 refresh 但不输出，fan_speed_name 等依赖它）
    trait_names = [
        "status", "dnd", "clean_summary", "sound_volume",
        "rooms", "maps", "consumables",
        "device_features",
    ]
    refresh_results = await asyncio.gather(
        *[_safe_refresh(getattr(props, name), name) for name in trait_names]
    )
    ok = dict(zip(trait_names, refresh_results))

    result: dict[str, Any] = {"device_name": device.name}

    # --- status trait ---
    if ok["status"]:
        status = props.status

        # 白名单字段
        for fname, desc in _STATUS_FIELDS.items():
            raw = getattr(status, fname, None)
            result[fname] = _field(_serialize_value(raw), desc)

        # computed properties — 用户友好值
        result["state"] = _field(status.state_name, "设备状态 (charging/cleaning/paused/returning_home/idle 等)")
        result["error_code"] = _field(status.error_code_name, "错误状态 (none=正常, lidar_blocked/bumper_stuck 等)")
        result["clean_area_m2"] = _field(status.square_meter_clean_area, "当前/最近清洁面积 (m²)")
        result["clean_time_minutes"] = _field(
            round(status.clean_time / 60, 1) if status.clean_time else 0,
            "当前/最近清洁时长 (分钟)",
        )
        result["fan_speed_name"] = _field(status.fan_speed_name, "吸力档位名称")
        result["water_mode_name"] = _field(status.water_mode_name, "拖布湿度名称")
        result["mop_route_name"] = _field(status.mop_route_name, "拖地路线名称")

        # dss 拆解（不输出原始 dss）
        result["clear_water_box_status"] = _field(
            _serialize_value(status.clear_water_box_status), "清水箱状态 (okay/out_of_water)")
        result["dirty_water_box_status"] = _field(
            _serialize_value(status.dirty_water_box_status), "污水箱状态 (okay/full_not_installed)")
        result["dust_bag_status"] = _field(
            _serialize_value(status.dust_bag_status), "尘袋状态 (okay/not_installed/full)")
        result["water_box_filter_status"] = _field(
            status.water_box_filter_status, "水滤网状态")
        result["clean_fluid_status"] = _field(
            _serialize_value(status.clean_fluid_status), "清洁液状态 (okay/empty_not_installed)")
        result["hatch_door_status"] = _field(
            status.hatch_door_status, "舱门状态")
        result["dock_cool_fan_status"] = _field(
            status.dock_cool_fan_status, "基座散热风扇状态")

    # --- dnd ---
    if ok["dnd"]:
        dnd = props.dnd
        result["dnd_enabled"] = _field(dnd.is_on, "勿扰模式是否启用")
        result["dnd_start_time"] = _field(
            f"{dnd.start_hour:02d}:{dnd.start_minute:02d}" if dnd.is_on else None,
            "勿扰开始时间 (HH:MM)",
        )
        result["dnd_end_time"] = _field(
            f"{dnd.end_hour:02d}:{dnd.end_minute:02d}" if dnd.is_on else None,
            "勿扰结束时间 (HH:MM)",
        )

    # --- clean_summary ---
    if ok["clean_summary"]:
        summary = props.clean_summary
        result["total_clean_time_hours"] = _field(
            round(summary.clean_time / 3600, 1) if summary.clean_time else 0,
            "累计清洁时长 (小时)",
        )
        result["total_clean_area_m2"] = _field(
            round(summary.clean_area / 1_000_000, 1) if summary.clean_area else 0,
            "累计清洁面积 (m²)",
        )
        result["total_clean_count"] = _field(summary.clean_count, "累计清洁次数")
        result["clean_record_count"] = _field(
            len(summary.records) if summary.records else 0, "清洁记录条数")

        if summary.last_clean_record:
            record = summary.last_clean_record
            result["last_clean_begin"] = _field(
                record.begin_datetime.isoformat() if record.begin_datetime else None,
                "最近清洁开始时间 (ISO8601)",
            )
            result["last_clean_end"] = _field(
                record.end_datetime.isoformat() if record.end_datetime else None,
                "最近清洁结束时间 (ISO8601)",
            )
            result["last_clean_duration_seconds"] = _field(
                record.duration, "最近清洁时长 (秒)")
            result["last_clean_area_m2"] = _field(
                record.area / 1_000_000 if record.area else 0,
                "最近清洁面积 (m²)",
            )
            result["last_clean_start_type"] = _field(
                record.start_type.name if record.start_type else None,
                "启动方式 (app/schedule/button/routines 等)",
            )
            result["last_clean_avoid_count"] = _field(
                record.avoid_count, "最近清洁避障次数")
            result["last_clean_wash_count"] = _field(
                record.wash_count, "最近清洁洗拖布次数")

    # --- sound_volume ---
    if ok["sound_volume"]:
        result["volume"] = _field(props.sound_volume.volume, "语音音量 (0-100)")

    # --- rooms ---
    if ok["rooms"]:
        rooms = props.rooms
        room_list = []
        for room in rooms.rooms or []:
            room_list.append({
                "segment_id": room.segment_id,
                "iot_id": room.iot_id,
                "name": room.raw_name or "<未命名>",
            })
        result["room_count"] = _field(len(room_list), "房间数量")
        result["rooms"] = _field(room_list, "房间列表 (含 segment_id 用于指定房间清洁)")

    # --- maps ---
    if ok["maps"]:
        maps = props.maps
        map_list = []
        for map_info in maps.map_info or []:
            map_list.append({
                "map_flag": map_info.map_flag,
                "name": map_info.name or "<未命名>",
                "room_count": len(map_info.rooms) if map_info.rooms else 0,
                "is_current": map_info.map_flag == maps.current_map,
            })
        result["current_map_flag"] = _field(maps.current_map, "当前地图 ID")
        result["map_count"] = _field(len(map_list), "地图数量")
        result["maps"] = _field(map_list, "地图列表")

    # --- consumables ---
    if ok["consumables"]:
        consumables = props.consumables
        SECS_PER_HOUR = 3600
        result["main_brush_hours"] = _field(
            round(consumables.main_brush_work_time / SECS_PER_HOUR, 1)
            if consumables.main_brush_work_time else 0,
            "主刷已使用时长 (小时，建议 300 小时更换)",
        )
        result["side_brush_hours"] = _field(
            round(consumables.side_brush_work_time / SECS_PER_HOUR, 1)
            if consumables.side_brush_work_time else 0,
            "边刷已使用时长 (小时，建议 200 小时更换)",
        )
        result["filter_hours"] = _field(
            round(consumables.filter_work_time / SECS_PER_HOUR, 1)
            if consumables.filter_work_time else 0,
            "滤网已使用时长 (小时，建议 150 小时更换)",
        )
        result["sensor_hours"] = _field(
            round(consumables.sensor_dirty_time / SECS_PER_HOUR, 1)
            if consumables.sensor_dirty_time else 0,
            "传感器清洁后使用时长 (小时，建议 30 小时清洁)",
        )

    return result


@mcp.tool(
    annotations={
        "title": "Get Map Content",
        "readOnlyHint": True,
        "destructiveHint": False,
        "idempotentHint": False,
        "openWorldHint": True,
    }
)
@_handle_errors
async def get_map_content(device_name: str | None = None) -> dict[str, Any]:
    """获取当前地图的元数据（通过 MQTT 拉取，数据量较大）。

    返回地图图片大小、机器人实时坐标、充电座坐标。不返回原始图片字节。
    此工具走 MQTT 拉取地图数据（几百KB~几MB），比 get_status 更重，
    仅在需要位置信息时调用。

    Args:
        device_name: 设备名称，支持模糊匹配。单设备时可省略。

    Returns:
        dict: {
            "device_name": str,
            "image_size_bytes": int,    # 地图图片大小
            "has_map_data": bool,       # 是否有地图数据
            "vacuum_position": {"x": int, "y": int},   # 机器人位置
            "charger_position": {"x": int, "y": int}    # 充电座位置
        }
    """
    conn = get_conn()
    device = await conn.get_device(device_name)

    if not device.v1_properties:
        return {"error": "设备不支持 V1 协议"}

    props = device.v1_properties
    await props.map_content.refresh()
    content = props.map_content

    result = {
        "device_name": device.name,
        "image_size_bytes": len(content.image_content) if content.image_content else 0,
        "has_map_data": content.map_data is not None,
    }

    if content.map_data:
        vacuum_pos = content.map_data.vacuum_position
        charger_pos = content.map_data.charger
        result["vacuum_position"] = {
            "x": vacuum_pos.x if vacuum_pos else None,
            "y": vacuum_pos.y if vacuum_pos else None,
        }
        result["charger_position"] = {
            "x": charger_pos.x if charger_pos else None,
            "y": charger_pos.y if charger_pos else None,
        }

    return result


# ============================================================================
# 控制类工具
# ============================================================================

@mcp.tool(
    annotations={
        "title": "Start Cleaning",
        "readOnlyHint": False,
        "destructiveHint": False,
        "idempotentHint": False,
        "openWorldHint": True,
    }
)
@_handle_errors
async def start_clean(
    device_name: str | None = None,
    segments: list[int] | None = None,
    repeat: int = 1,
) -> dict[str, Any]:
    """开始清洁。

    不传 segments 则全屋清洁，传 segments 则清洁指定房间。
    房间 segment_id 可通过 get_status 的 rooms 字段获取。
    命令已发送并确认，但设备执行需要几秒，建议稍后调用 get_status 查看最新状态。

    Args:
        device_name: 设备名称，支持模糊匹配。单设备时可省略。
        segments: 房间 segment_id 列表。不传则全屋清洁。
        repeat: 清洁遍数，默认 1。
    """
    conn = get_conn()
    device = await conn.get_device(device_name)
    if not device.v1_properties:
        return {"error": "设备不支持 V1 协议"}

    cmd = device.v1_properties.command
    if segments:
        params = [{"segments": segments, "repeat": repeat}]
        await cmd.send(RoborockCommand.APP_SEGMENT_CLEAN, params=params)
        return {"status": "ok", "action": "segment_clean", "segments": segments, "repeat": repeat}
    else:
        await cmd.send(RoborockCommand.APP_START)
        return {"status": "ok", "action": "full_clean"}


@mcp.tool(
    annotations={
        "title": "Stop Cleaning",
        "readOnlyHint": False,
        "destructiveHint": False,
        "idempotentHint": True,
        "openWorldHint": True,
    }
)
@_handle_errors
async def stop_clean(device_name: str | None = None) -> dict[str, Any]:
    """停止清洁。机器人将原地停止。

    命令已发送并确认，但设备执行需要几秒，建议稍后调用 get_status 查看最新状态。

    Args:
        device_name: 设备名称，支持模糊匹配。单设备时可省略。
    """
    conn = get_conn()
    device = await conn.get_device(device_name)
    if not device.v1_properties:
        return {"error": "设备不支持 V1 协议"}

    await device.v1_properties.command.send(RoborockCommand.APP_STOP)
    return {"status": "ok", "action": "stop"}


@mcp.tool(
    annotations={
        "title": "Pause Cleaning",
        "readOnlyHint": False,
        "destructiveHint": False,
        "idempotentHint": True,
        "openWorldHint": True,
    }
)
@_handle_errors
async def pause_clean(device_name: str | None = None) -> dict[str, Any]:
    """暂停清洁。可通过 resume_clean 恢复。

    命令已发送并确认，但设备执行需要几秒，建议稍后调用 get_status 查看最新状态。

    Args:
        device_name: 设备名称，支持模糊匹配。单设备时可省略。
    """
    conn = get_conn()
    device = await conn.get_device(device_name)
    if not device.v1_properties:
        return {"error": "设备不支持 V1 协议"}

    await device.v1_properties.command.send(RoborockCommand.APP_PAUSE)
    return {"status": "ok", "action": "pause"}


@mcp.tool(
    annotations={
        "title": "Resume Cleaning",
        "readOnlyHint": False,
        "destructiveHint": False,
        "idempotentHint": True,
        "openWorldHint": True,
    }
)
@_handle_errors
async def resume_clean(device_name: str | None = None) -> dict[str, Any]:
    """恢复之前暂停的清洁任务。

    使用 APP_START 恢复，无论全屋还是分区清洁均可恢复。
    命令已发送并确认，但设备执行需要几秒，建议稍后调用 get_status 查看最新状态。

    Args:
        device_name: 设备名称，支持模糊匹配。单设备时可省略。
    """
    conn = get_conn()
    device = await conn.get_device(device_name)
    if not device.v1_properties:
        return {"error": "设备不支持 V1 协议"}

    await device.v1_properties.command.send(RoborockCommand.APP_START)
    return {"status": "ok", "action": "resume"}


@mcp.tool(
    annotations={
        "title": "Go Home",
        "readOnlyHint": False,
        "destructiveHint": False,
        "idempotentHint": True,
        "openWorldHint": True,
    }
)
@_handle_errors
async def go_home(device_name: str | None = None) -> dict[str, Any]:
    """让机器人回充电座。

    命令已发送并确认，但设备执行需要几秒，建议稍后调用 get_status 查看最新状态。

    Args:
        device_name: 设备名称，支持模糊匹配。单设备时可省略。
    """
    conn = get_conn()
    device = await conn.get_device(device_name)
    if not device.v1_properties:
        return {"error": "设备不支持 V1 协议"}

    await device.v1_properties.command.send(RoborockCommand.APP_CHARGE)
    return {"status": "ok", "action": "go_home"}


@mcp.tool(
    annotations={
        "title": "Find Robot",
        "readOnlyHint": False,
        "destructiveHint": False,
        "idempotentHint": True,
        "openWorldHint": True,
    }
)
@_handle_errors
async def find_robot(device_name: str | None = None) -> dict[str, Any]:
    """让机器人发出声音，帮助定位。

    命令已发送并确认，但设备执行需要几秒。

    Args:
        device_name: 设备名称，支持模糊匹配。单设备时可省略。
    """
    conn = get_conn()
    device = await conn.get_device(device_name)
    if not device.v1_properties:
        return {"error": "设备不支持 V1 协议"}

    await device.v1_properties.command.send(RoborockCommand.FIND_ME)
    return {"status": "ok", "action": "find_robot"}


@mcp.tool(
    annotations={
        "title": "Dock Action",
        "readOnlyHint": False,
        "destructiveHint": False,
        "idempotentHint": False,
        "openWorldHint": True,
    }
)
@_handle_errors
async def dock_action(
    action: str,
    device_name: str | None = None,
) -> dict[str, Any]:
    """控制基座操作。

    命令已发送并确认，但设备执行需要几秒，建议稍后调用 get_status 查看最新状态。

    Args:
        action: 操作类型。可选值：
            - "start_collect_dust": 开始集尘
            - "stop_collect_dust": 停止集尘
            - "start_wash": 开始洗拖布
            - "stop_wash": 停止洗拖布
        device_name: 设备名称，支持模糊匹配。单设备时可省略。
    """
    ACTION_MAP = {
        "start_collect_dust": RoborockCommand.APP_START_COLLECT_DUST,
        "stop_collect_dust": RoborockCommand.APP_STOP_COLLECT_DUST,
        "start_wash": RoborockCommand.APP_START_WASH,
        "stop_wash": RoborockCommand.APP_STOP_WASH,
    }
    if action not in ACTION_MAP:
        return {"error": f"未知操作: {action}，可选: {', '.join(ACTION_MAP)}"}

    conn = get_conn()
    device = await conn.get_device(device_name)
    if not device.v1_properties:
        return {"error": "设备不支持 V1 协议"}

    await device.v1_properties.command.send(ACTION_MAP[action])
    return {"status": "ok", "action": action}


@mcp.tool(
    annotations={
        "title": "Set Volume",
        "readOnlyHint": False,
        "destructiveHint": False,
        "idempotentHint": True,
        "openWorldHint": True,
    }
)
@_handle_errors
async def set_volume(
    device_name: str | None = None,
    volume: int = 50,
) -> dict[str, Any]:
    """设置机器人语音音量。

    命令已发送并确认，但设备执行需要几秒，建议稍后调用 get_status 查看最新状态。

    Args:
        device_name: 设备名称，支持模糊匹配。单设备时可省略。
        volume: 音量值，0-100。
    """
    if not 0 <= volume <= 100:
        return {"error": "音量值需在 0-100 之间"}

    conn = get_conn()
    device = await conn.get_device(device_name)
    if not device.v1_properties:
        return {"error": "设备不支持 V1 协议"}

    await device.v1_properties.sound_volume.set_volume(volume)
    return {"status": "ok", "volume": volume}


@mcp.tool(
    annotations={
        "title": "List Routines",
        "readOnlyHint": True,
        "destructiveHint": False,
        "idempotentHint": True,
        "openWorldHint": True,
    }
)
@_handle_errors
async def get_routines(device_name: str | None = None) -> dict[str, Any]:
    """列出设备的智能场景（通过云端 API 获取）。

    返回可执行的预设场景列表，每个场景有 id 和名称。
    使用 execute_routine 执行具体场景。

    Args:
        device_name: 设备名称，支持模糊匹配。单设备时可省略。
    """
    conn = get_conn()
    device = await conn.get_device(device_name)
    if not device.v1_properties:
        return {"error": "设备不支持 V1 协议"}

    routines = await device.v1_properties.routines.get_routines()
    return {
        "device_name": device.name,
        "count": len(routines),
        "routines": [{"id": r.id, "name": r.name} for r in routines],
    }


@mcp.tool(
    annotations={
        "title": "Execute Routine",
        "readOnlyHint": False,
        "destructiveHint": False,
        "idempotentHint": False,
        "openWorldHint": True,
    }
)
@_handle_errors
async def execute_routine(
    routine_id: int,
    device_name: str | None = None,
) -> dict[str, Any]:
    """执行智能场景。

    通过 get_routines 获取可用场景及其 ID，然后传入 routine_id 执行。
    命令已发送并确认，但设备执行需要几秒，建议稍后调用 get_status 查看最新状态。

    Args:
        routine_id: 场景 ID，通过 get_routines 获取。必填参数。
        device_name: 设备名称，支持模糊匹配。单设备时可省略。
    """
    conn = get_conn()
    device = await conn.get_device(device_name)
    if not device.v1_properties:
        return {"error": "设备不支持 V1 协议"}

    await device.v1_properties.routines.execute_routine(routine_id)
    return {"status": "ok", "routine_id": routine_id}



"""Roborock V1 Protocol Encoder."""

from __future__ import annotations

import base64
import json
import logging
import secrets
import struct
from collections.abc import Callable
from dataclasses import dataclass, field
from enum import StrEnum
from typing import Any, Protocol, TypeVar, overload

from roborock.data import RoborockBase, RRiot
from roborock.exceptions import RoborockException, RoborockInvalidStatus, RoborockUnsupportedFeature
from roborock.protocol import Utils
from roborock.roborock_message import RoborockMessage, RoborockMessageProtocol
from roborock.roborock_typing import RoborockCommand
from roborock.util import get_next_int, get_timestamp

_LOGGER = logging.getLogger(__name__)

__all__ = [
    "SecurityData",
    "create_security_data",
    "decode_rpc_response",
    "V1RpcChannel",
]

CommandType = RoborockCommand | str
ParamsType = list | dict | int | None


class LocalProtocolVersion(StrEnum):
    """Supported local protocol versions. Different from vacuum protocol versions."""

    L01 = "L01"
    V1 = "1.0"


@dataclass(frozen=True, kw_only=True)
class SecurityData:
    """Security data included in the request for some V1 commands."""

    endpoint: str
    nonce: bytes

    def to_dict(self) -> dict[str, Any]:
        """Convert security data to a dictionary for sending in the payload."""
        return {"security": {"endpoint": self.endpoint, "nonce": self.nonce.hex().lower()}}

    def to_diagnostic_data(self) -> dict[str, Any]:
        """Convert security data to a dictionary for debugging purposes."""
        return {"nonce": self.nonce.hex().lower()}


def create_security_data(rriot: RRiot) -> SecurityData:
    """Create a SecurityData instance for the given endpoint and nonce."""
    nonce = secrets.token_bytes(16)
    endpoint = base64.b64encode(Utils.md5(rriot.k.encode())[8:14]).decode()
    return SecurityData(endpoint=endpoint, nonce=nonce)


@dataclass
class RequestMessage:
    """Data structure for v1 RoborockMessage payloads."""

    method: RoborockCommand | str
    params: ParamsType
    timestamp: int = field(default_factory=lambda: get_timestamp())
    request_id: int = field(default_factory=lambda: get_next_int(10000, 32767))

    def encode_message(
        self,
        protocol: RoborockMessageProtocol,
        security_data: SecurityData | None = None,
        version: LocalProtocolVersion = LocalProtocolVersion.V1,
    ) -> RoborockMessage:
        """Convert the request message to a RoborockMessage."""
        return RoborockMessage(
            timestamp=self.timestamp,
            protocol=protocol,
            payload=self._as_payload(security_data=security_data),
            version=version.value.encode(),
        )

    def _as_payload(self, security_data: SecurityData | None) -> bytes:
        """Convert the request arguments to a dictionary."""
        inner = {
            "id": self.request_id,
            "method": self.method,
            "params": self.params or [],
            **(security_data.to_dict() if security_data else {}),
        }
        return bytes(
            json.dumps(
                {
                    "dps": {"101": json.dumps(inner, separators=(",", ":"))},
                    "t": self.timestamp,
                },
                separators=(",", ":"),
            ).encode()
        )


ResponseData = dict[str, Any] | list | int

# V1 RPC error code mappings to specific exception types
_V1_ERROR_CODE_EXCEPTIONS: dict[int, type[RoborockException]] = {
    -10007: RoborockInvalidStatus,  # "invalid status" - device action locked
}


def _create_api_error(error: Any) -> RoborockException:
    """Create an appropriate exception for a V1 RPC error response.

    Maps known error codes to specific exception types for easier handling
    at higher levels.
    """
    if isinstance(error, dict):
        code = error.get("code")
        if isinstance(code, int) and (exc_type := _V1_ERROR_CODE_EXCEPTIONS.get(code)):
            return exc_type(error)
    return RoborockException(error)


@dataclass(kw_only=True, frozen=True)
class ResponseMessage:
    """Data structure for v1 RoborockMessage responses."""

    request_id: int | None
    """The request ID of the response."""

    data: ResponseData
    """The data of the response, where the type depends on the command."""

    api_error: RoborockException | None = None
    """The API error message of the response if any."""


def decode_rpc_response(message: RoborockMessage) -> ResponseMessage:
    """Decode a V1 RPC_RESPONSE message.

    This will raise a RoborockException if the message cannot be parsed. A
    response object will be returned even if there is an error in the
    response, as long as we can extract the request ID. This is so we can
    associate an API response with a request even if there was an error.
    """
    if not message.payload:
        return ResponseMessage(request_id=message.seq, data={})
    try:
        payload = json.loads(message.payload.decode())
    except (json.JSONDecodeError, TypeError, UnicodeDecodeError) as e:
        raise RoborockException(f"Invalid V1 message payload: {e} for {message.payload!r}") from e

    _LOGGER.debug("Decoded V1 message payload: %s", payload)
    datapoints = payload.get("dps", {})
    if not isinstance(datapoints, dict):
        raise RoborockException(f"Invalid V1 message format: 'dps' should be a dictionary for {message.payload!r}")

    if not (data_point := datapoints.get(str(RoborockMessageProtocol.RPC_RESPONSE))):
        raise RoborockException(
            f"Invalid V1 message format: missing '{RoborockMessageProtocol.RPC_RESPONSE}' data point"
        )

    try:
        data_point_response = json.loads(data_point)
    except (json.JSONDecodeError, TypeError) as e:
        raise RoborockException(
            f"Invalid V1 message data point '{RoborockMessageProtocol.RPC_RESPONSE}': {e} for {message.payload!r}"
        ) from e

    request_id: int | None = data_point_response.get("id")
    api_error: RoborockException | None = None
    if error := data_point_response.get("error"):
        api_error = _create_api_error(error)

    if (result := data_point_response.get("result")) is None:
        # Some firmware versions return an error-only response (no "result" key).
        # Preserve that error instead of overwriting it with a parsing exception.
        if api_error is None:
            api_error = RoborockException(
                f"Invalid V1 message format: missing 'result' in data point for {message.payload!r}"
            )
        result = {}
    else:
        _LOGGER.debug("Decoded V1 message result: %s", result)
        if isinstance(result, str):
            if result == "unknown_method":
                api_error = RoborockUnsupportedFeature("The method called is not recognized by the device.")
            elif result != "ok":
                api_error = RoborockException(f"Unexpected API Result: {result}")
            result = {}
        if not isinstance(result, dict | list | int):
            # If we already have an API error, prefer returning a response object
            # rather than failing to decode the message entirely.
            if api_error is None:
                raise RoborockException(
                    f"Invalid V1 message format: 'result' was unexpected type {type(result)}. {message.payload!r}"
                )
            result = {}

    if not request_id and api_error:
        raise api_error
    return ResponseMessage(request_id=request_id, data=result, api_error=api_error)


@dataclass
class MapResponse:
    """Data structure for the V1 Map response."""

    request_id: int
    """The request ID of the map response."""

    data: bytes
    """The map data, decrypted and decompressed."""


def create_map_response_decoder(security_data: SecurityData) -> Callable[[RoborockMessage], MapResponse | None]:
    """Create a decoder for V1 map response messages."""

    def _decode_map_response(message: RoborockMessage) -> MapResponse | None:
        """Decode a V1 map response message."""
        if not message.payload or len(message.payload) < 24:
            raise RoborockException("Invalid V1 map response format: missing payload")
        header, body = message.payload[:24], message.payload[24:]
        [endpoint, _, request_id, _] = struct.unpack("<8s8sH6s", header)
        if not endpoint.decode().startswith(security_data.endpoint):
            _LOGGER.debug("Received map response not requested by this device, ignoring.")
            return None
        try:
            decrypted = Utils.decrypt_cbc(body, security_data.nonce)
        except ValueError as err:
            raise RoborockException("Failed to decode map message payload") from err
        decompressed = Utils.decompress(decrypted)
        return MapResponse(request_id=request_id, data=decompressed)

    return _decode_map_response


_T = TypeVar("_T", bound=RoborockBase)


class V1RpcChannel(Protocol):
    """Protocol for V1 RPC channels.

    This is a wrapper around a raw channel that provides a high-level interface
    for sending commands and receiving responses.
    """

    @overload
    async def send_command(
        self,
        method: CommandType,
        *,
        params: ParamsType = None,
    ) -> Any:
        """Send a command and return a decoded response."""
        ...

    @overload
    async def send_command(
        self,
        method: CommandType,
        *,
        response_type: type[_T],
        params: ParamsType = None,
    ) -> _T:
        """Send a command and return a parsed response RoborockBase type."""
        ...



from roborock.data import LedStatus
from roborock.devices.traits.v1 import common
from roborock.roborock_typing import RoborockCommand

from .common import V1ResponseData


class LedStatusTrait(LedStatus, common.V1TraitMixin, common.RoborockSwitchBase):
    """Trait for controlling the LED status of a Roborock device."""

    command = RoborockCommand.GET_LED_STATUS
    requires_feature = "is_led_status_switch_supported"

    @property
    def is_on(self) -> bool:
        """Return whether the LED status is enabled."""
        return self.status == 1

    async def enable(self) -> None:
        """Enable the LED status."""
        await self.rpc_channel.send_command(RoborockCommand.SET_LED_STATUS, params=[1])
        # Optimistic update to avoid an extra refresh
        self.status = 1

    async def disable(self) -> None:
        """Disable the LED status."""
        await self.rpc_channel.send_command(RoborockCommand.SET_LED_STATUS, params=[0])
        # Optimistic update to avoid an extra refresh
        self.status = 0

    @classmethod
    def _parse_type_response(cls, response: V1ResponseData) -> LedStatus:
        """Parse the response from the device into a a RoborockBase.

        Subclasses should override this method to implement custom parsing
        logic as needed.
        """
        if not isinstance(response, list):
            raise ValueError(f"Unexpected {cls} response format: {response!r}")
        response = response[0]
        if not isinstance(response, int):
            raise ValueError(f"Unexpected {cls} response format: {response!r}")
        return cls.from_dict({"status": response})



"""Module for device traits.

This package contains the trait definitions for different device protocols supported
by Roborock devices.

Submodules
----------
*   `v1`: Contains traits for standard Roborock vacuums (e.g., S-series, Q-series).
    These devices use the V1 protocol and have rich feature sets split into
    granular traits (e.g., `StatusTrait`, `ConsumableTrait`).
*   `a01`: Contains APIs for A01 protocol devices, such as the Dyad (wet/dry vacuum)
    and Zeo (washing machine). These devices use a different communication structure.
*   `b01`: Contains APIs for B01 protocol devices.
"""

from abc import ABC

__all__ = [
    "Trait",
    "traits_mixin",
    "v1",
    "a01",
    "b01",
]


class Trait(ABC):
    """Base class for all traits."""



import dataclasses
import datetime
import inspect
import json
import logging
import re
import types
from dataclasses import asdict, dataclass, field
from enum import Enum
from functools import cached_property
from typing import Any, ClassVar, NamedTuple, get_args, get_origin

from .code_mappings import (
    SHORT_MODEL_TO_ENUM,
    RoborockCategory,
    RoborockModeEnum,
    RoborockProductNickname,
)

_LOGGER = logging.getLogger(__name__)


def _camelize(s: str):
    first, *others = s.split("_")
    if len(others) == 0:
        return s
    return "".join([first.lower(), *map(str.title, others)])


def _decamelize(s: str):
    # Split before uppercase letters not at the start, and before numbers
    s = re.sub(r"(?<=[a-z0-9])([A-Z])", r"_\1", s)
    s = re.sub(r"([A-Z]+)([A-Z][a-z])", r"\1_\2", s)  # Split acronyms followed by normal camelCase
    s = re.sub(r"([a-zA-Z])([0-9]+)", r"\1_\2", s)
    s = s.lower()
    # Temporary fix to avoid breaking any serialization.
    s = s.replace("base_64", "base64")
    return s


def _attr_repr(obj: Any) -> str:
    """Return a string representation of the object including specified attributes.

    This reproduces the default repr behavior of dataclasses, but also includes
    properties. This must be called by the child class's __repr__ method since
    the parent RoborockBase class does not know about the child class's attributes.
    """
    # Reproduce default repr behavior
    parts = []
    for k in dir(obj):
        if k.startswith("_"):
            continue
        try:
            v = getattr(obj, k)
        except (RuntimeError, Exception):
            continue
        if callable(v):
            continue
        parts.append(f"{k}={v!r}")
    return f"{type(obj).__name__}({', '.join(parts)})"


@dataclass(repr=False)
class RoborockBase:
    """Base class for all Roborock data classes."""

    _missing_logged: ClassVar[set[str]] = set()

    @staticmethod
    def _convert_to_class_obj(class_type: type, value):
        if get_origin(class_type) is list:
            sub_type = get_args(class_type)[0]
            return [RoborockBase._convert_to_class_obj(sub_type, obj) for obj in value]
        if get_origin(class_type) is dict:
            key_type, value_type = get_args(class_type)
            if key_type is not None:
                return {key_type(k): RoborockBase._convert_to_class_obj(value_type, v) for k, v in value.items()}
            return {k: RoborockBase._convert_to_class_obj(value_type, v) for k, v in value.items()}
        if inspect.isclass(class_type):
            if issubclass(class_type, RoborockBase):
                return class_type.from_dict(value)
            if issubclass(class_type, RoborockModeEnum):
                return class_type.from_code(value)
        if class_type is Any or type(class_type) is str:
            return value
        return class_type(value)  # type: ignore[call-arg]

    @classmethod
    def from_dict(cls, data: dict[str, Any]):
        """Create an instance of the class from a dictionary."""
        if not isinstance(data, dict):
            return None
        field_types = {field.name: field.type for field in dataclasses.fields(cls)}
        normalized_data: dict[str, Any] = {}
        for orig_key, value in data.items():
            key = _decamelize(orig_key)
            if field_types.get(key) is None:
                if (log_key := f"{cls.__name__}.{key}") not in RoborockBase._missing_logged:
                    _LOGGER.debug(
                        "Key '%s' (decamelized: '%s') not found in %s fields, skipping",
                        orig_key,
                        key,
                        cls.__name__,
                    )
                    RoborockBase._missing_logged.add(log_key)
                continue
            normalized_data[key] = value

        result = RoborockBase.convert_dict(field_types, normalized_data)
        return cls(**result)

    @staticmethod
    def convert_dict(types_map: dict[Any, type], data: dict[Any, Any]) -> dict[Any, Any]:
        """Generic helper to convert a dictionary of values based on a schema map of types.

        This is meant to be used by traits that use dataclass reflection similar to
        `Roborock.from_dict` to merge in new data updates.
        """
        result: dict[Any, Any] = {}
        for key, value in data.items():
            if key not in types_map:
                continue
            field_type = types_map[key]
            if value == "None" or value is None:
                result[key] = None
                continue
            if isinstance(field_type, types.UnionType):
                for subtype in get_args(field_type):
                    if subtype is types.NoneType:
                        continue
                    try:
                        result[key] = RoborockBase._convert_to_class_obj(subtype, value)
                        break
                    except Exception:
                        _LOGGER.exception(f"Failed to convert {key} with value {value} to type {subtype}")
                        continue
            else:
                try:
                    result[key] = RoborockBase._convert_to_class_obj(field_type, value)
                except Exception:
                    _LOGGER.exception(f"Failed to convert {key} with value {value} to type {field_type}")
                    continue

        return result

    def as_dict(self) -> dict:
        return asdict(
            self,
            dict_factory=lambda _fields: {
                _camelize(key): value.value if isinstance(value, Enum) else value
                for (key, value) in _fields
                if value is not None
            },
        )


@dataclass
class RoborockBaseTimer(RoborockBase):
    start_hour: int | None = None
    start_minute: int | None = None
    end_hour: int | None = None
    end_minute: int | None = None
    enabled: int | None = None

    @property
    def start_time(self) -> datetime.time | None:
        return (
            datetime.time(hour=self.start_hour, minute=self.start_minute)
            if self.start_hour is not None and self.start_minute is not None
            else None
        )

    @property
    def end_time(self) -> datetime.time | None:
        return (
            datetime.time(hour=self.end_hour, minute=self.end_minute)
            if self.end_hour is not None and self.end_minute is not None
            else None
        )

    def as_list(self) -> list:
        return [self.start_hour, self.start_minute, self.end_hour, self.end_minute]

    def __repr__(self) -> str:
        return _attr_repr(self)


@dataclass
class Reference(RoborockBase):
    r: str | None = None
    a: str | None = None
    m: str | None = None
    l: str | None = None


@dataclass
class RRiot(RoborockBase):
    u: str
    s: str
    h: str
    k: str
    r: Reference


@dataclass
class UserData(RoborockBase):
    rriot: RRiot
    uid: int | None = None
    tokentype: str | None = None
    token: str | None = None
    rruid: str | None = None
    region: str | None = None
    countrycode: str | None = None
    country: str | None = None
    nickname: str | None = None
    tuya_device_state: int | None = None
    avatarurl: str | None = None


@dataclass
class HomeDataProductSchema(RoborockBase):
    id: Any | None = None
    name: Any | None = None
    code: Any | None = None
    mode: Any | None = None
    type: Any | None = None
    product_property: Any | None = None
    property: Any | None = None
    desc: Any | None = None


@dataclass
class HomeDataProduct(RoborockBase):
    id: str
    name: str
    model: str
    category: RoborockCategory
    code: str | None = None
    icon_url: str | None = None
    attribute: Any | None = None
    capability: int | None = None
    schema: list[HomeDataProductSchema] | None = None

    @property
    def product_nickname(self) -> RoborockProductNickname:
        return SHORT_MODEL_TO_ENUM.get(self.model.split(".")[-1], RoborockProductNickname.PEARLPLUS)

    def summary_info(self) -> str:
        """Return a string with key product information for logging purposes."""
        return f"{self.name} (model={self.model}, category={self.category})"

    @cached_property
    def supported_schema_codes(self) -> set[str]:
        """Return a set of fields that are supported by the device."""
        if self.schema is None:
            return set()
        return {schema.code for schema in self.schema if schema.code is not None}


@dataclass
class HomeDataDevice(RoborockBase):
    duid: str
    name: str
    local_key: str
    product_id: str
    fv: str | None = None
    attribute: Any | None = None
    active_time: int | None = None
    runtime_env: Any | None = None
    time_zone_id: str | None = None
    icon_url: str | None = None
    lon: Any | None = None
    lat: Any | None = None
    share: Any | None = None
    share_time: Any | None = None
    online: bool | None = None
    pv: str | None = None
    room_id: Any | None = None
    tuya_uuid: Any | None = None
    tuya_migrated: bool | None = None
    extra: Any | None = None
    sn: str | None = None
    feature_set: str | None = None
    new_feature_set: str | None = None
    device_status: dict | None = None
    silent_ota_switch: bool | None = None
    setting: Any | None = None
    f: bool | None = None
    create_time: int | None = None
    cid: str | None = None
    share_type: Any | None = None
    share_expired_time: int | None = None

    def summary_info(self) -> str:
        """Return a string with key device information for logging purposes."""
        return f"{self.name} (pv={self.pv}, fv={self.fv}, online={self.online})"


@dataclass
class HomeDataRoom(RoborockBase):
    id: int
    name: str

    @property
    def iot_id(self) -> str:
        """Return the room's ID as a string IOT ID."""
        return str(self.id)


@dataclass
class HomeDataScene(RoborockBase):
    id: int
    name: str


@dataclass
class HomeDataSchedule(RoborockBase):
    id: int
    cron: str
    repeated: bool
    enabled: bool
    param: dict | None = None


@dataclass
class HomeData(RoborockBase):
    id: int
    name: str
    products: list[HomeDataProduct] = field(default_factory=lambda: [])
    devices: list[HomeDataDevice] = field(default_factory=lambda: [])
    received_devices: list[HomeDataDevice] = field(default_factory=lambda: [])
    lon: Any | None = None
    lat: Any | None = None
    geo_name: Any | None = None
    rooms: list[HomeDataRoom] = field(default_factory=list)

    def get_all_devices(self) -> list[HomeDataDevice]:
        devices = []
        if self.devices is not None:
            devices += self.devices
        if self.received_devices is not None:
            devices += self.received_devices
        return devices

    @cached_property
    def product_map(self) -> dict[str, HomeDataProduct]:
        """Returns a dictionary of product IDs to HomeDataProduct objects."""
        return {product.id: product for product in self.products}

    @cached_property
    def device_products(self) -> dict[str, tuple[HomeDataDevice, HomeDataProduct]]:
        """Returns a dictionary of device DUIDs to HomeDataDeviceProduct objects."""
        product_map = self.product_map
        return {
            device.duid: (device, product)
            for device in self.get_all_devices()
            if (product := product_map.get(device.product_id)) is not None
        }

    @property
    def rooms_map(self) -> dict[str, HomeDataRoom]:
        """Returns a dictionary of Room iot_id to rooms"""
        return {room.iot_id: room for room in self.rooms}

    @property
    def rooms_name_map(self) -> dict[str, str]:
        """Returns a dictionary of Room iot_id to room names."""
        return {room.iot_id: room.name for room in self.rooms}


@dataclass
class LoginData(RoborockBase):
    user_data: UserData
    email: str
    home_data: HomeData | None = None


@dataclass
class DeviceData(RoborockBase):
    device: HomeDataDevice
    model: str
    host: str | None = None

    @property
    def product_nickname(self) -> RoborockProductNickname:
        return SHORT_MODEL_TO_ENUM.get(self.model.split(".")[-1], RoborockProductNickname.PEARLPLUS)

    def __repr__(self) -> str:
        return _attr_repr(self)


@dataclass
class RoomMapping(RoborockBase):
    segment_id: int
    iot_id: str


@dataclass
class NamedRoomMapping(RoomMapping):
    """Dataclass representing a mapping of a room segment to a name.

    The name information is not provided by the device directly, but is provided
    from the HomeData based on the iot_id from the room.
    """

    @property
    def name(self) -> str:
        """The human-readable name of the room, or a default name if not available."""
        return self.raw_name or f"Room {self.segment_id}"

    raw_name: str | None = None
    """The raw name of the room, as provided by the device."""


@dataclass
class CombinedMapInfo(RoborockBase):
    """Data structure for caching home information.

    This is not provided directly by the API, but is a combination of map data
    and room data to provide a more useful structure.
    """

    map_flag: int
    """The map identifier."""

    name: str
    """The name of the map from MultiMapsListMapInfo."""

    rooms: list[NamedRoomMapping]
    """The list of rooms in the map."""

    @property
    def rooms_map(self) -> dict[int, NamedRoomMapping]:
        """Returns a mapping of segment_id to NamedRoomMapping."""
        return {room.segment_id: room for room in self.rooms}


@dataclass
class BroadcastMessage(RoborockBase):
    duid: str
    ip: str
    version: bytes


class ServerTimer(NamedTuple):
    id: str
    status: str
    dontknow: int


@dataclass
class RoborockProductStateValue(RoborockBase):
    value: list
    desc: dict


@dataclass
class RoborockProductState(RoborockBase):
    dps: int
    desc: dict
    value: list[RoborockProductStateValue]


@dataclass
class RoborockProductSpec(RoborockBase):
    state: RoborockProductState
    battery: dict | None = None
    dry_countdown: dict | None = None
    extra: dict | None = None
    offpeak: dict | None = None
    countdown: dict | None = None
    mode: dict | None = None
    ota_nfo: dict | None = None
    pause: dict | None = None
    program: dict | None = None
    shutdown: dict | None = None
    washing_left: dict | None = None


@dataclass
class RoborockProduct(RoborockBase):
    id: int | None = None
    name: str | None = None
    model: str | None = None
    packagename: str | None = None
    ssid: str | None = None
    picurl: str | None = None
    cardpicurl: str | None = None
    mediumCardpicurl: str | None = None
    resetwifipicurl: str | None = None
    configPicUrl: str | None = None
    pluginPicUrl: str | None = None
    resetwifitext: dict | None = None
    tuyaid: str | None = None
    status: int | None = None
    rriotid: str | None = None
    pictures: list | None = None
    ncMode: str | None = None
    scope: str | None = None
    product_tags: list | None = None
    agreements: list | None = None
    cardspec: str | None = None
    plugin_pic_url: str | None = None

    @property
    def product_nickname(self) -> RoborockProductNickname | None:
        if self.cardspec:
            return RoborockProductSpec.from_dict(json.loads(self.cardspec).get("data"))
        return None

    def __repr__(self) -> str:
        return _attr_repr(self)


@dataclass
class RoborockProductCategory(RoborockBase):
    id: int
    display_name: str
    icon_url: str


@dataclass
class RoborockCategoryDetail(RoborockBase):
    category: RoborockProductCategory
    product_list: list[RoborockProduct]


@dataclass
class ProductResponse(RoborockBase):
    category_detail_list: list[RoborockCategoryDetail]



"""Roborock A01 Protocol encoding and decoding."""

import json
import logging
from collections.abc import Callable
from typing import Any

from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad

from roborock.exceptions import RoborockException
from roborock.roborock_message import (
    RoborockDyadDataProtocol,
    RoborockMessage,
    RoborockMessageProtocol,
    RoborockZeoProtocol,
)

_LOGGER = logging.getLogger(__name__)

A01_VERSION = b"A01"


def _no_encode(value: Any) -> Any:
    return value


def encode_mqtt_payload(
    data: dict[RoborockDyadDataProtocol, Any]
    | dict[RoborockZeoProtocol, Any]
    | dict[RoborockDyadDataProtocol | RoborockZeoProtocol, Any],
    value_encoder: Callable[[Any], Any] | None = None,
) -> RoborockMessage:
    """Encode payload for A01 commands over MQTT.

    Args:
        data: The data to encode.
        value_encoder: A function to encode the values of the dictionary.

    Returns:
        RoborockMessage: The encoded message.
    """
    if value_encoder is None:
        value_encoder = _no_encode
    dps_data = {"dps": {key: value_encoder(value) for key, value in data.items()}}
    payload = pad(json.dumps(dps_data).encode("utf-8"), AES.block_size)
    return RoborockMessage(
        protocol=RoborockMessageProtocol.RPC_REQUEST,
        version=A01_VERSION,
        payload=payload,
    )


def decode_rpc_response(message: RoborockMessage) -> dict[int, Any]:
    """Decode a V1 RPC_RESPONSE message."""
    if not message.payload:
        raise RoborockException("Invalid A01 message format: missing payload")
    try:
        unpadded = unpad(message.payload, AES.block_size)
    except ValueError as err:
        raise RoborockException(f"Unable to unpad A01 payload: {err}")

    try:
        payload = json.loads(unpadded.decode())
    except (json.JSONDecodeError, TypeError) as e:
        raise RoborockException(f"Invalid A01 message payload: {e} for {message.payload!r}") from e

    datapoints = payload.get("dps", {})
    if not isinstance(datapoints, dict):
        raise RoborockException(f"Invalid A01 message format: 'dps' should be a dictionary for {message.payload!r}")
    try:
        return {int(key): value for key, value in datapoints.items()}
    except ValueError:
        raise RoborockException(f"Invalid A01 message format: 'dps' key should be an integer for {message.payload!r}")



from ..code_mappings import RoborockModeEnum


class B01_Q10_DP(RoborockModeEnum):
    CLEAN_TIME = ("dpCleanTime", 6)
    CLEAN_AREA = ("dpCleanArea", 7)
    SEEK = ("dpSeek", 11)
    REMOTE = ("dpRemote", 12)
    MAP_RESET = ("dpMapReset", 13)
    REQUEST = ("dpRequest", 16)
    RESET_SIDE_BRUSH = ("dpResetSideBrush", 18)
    RESET_MAIN_BRUSH = ("dpResetMainBrush", 20)
    RESET_FILTER = ("dpResetFilter", 22)
    RAG_LIFE = ("dpRagLife", 23)
    RESET_RAG_LIFE = ("dpResetRagLife", 24)
    NOT_DISTURB = ("dpNotDisturb", 25)
    VOLUME = ("dpVolume", 26)
    BEAK_CLEAN = ("dpBeakClean", 27)
    TOTAL_CLEAN_AREA = ("dpTotalCleanArea", 29)
    TOTAL_CLEAN_COUNT = ("dpTotalCleanCount", 30)
    TOTAL_CLEAN_TIME = ("dpTotalCleanTime", 31)
    TIMER = ("dpTimer", 32)
    NOT_DISTURB_DATA = ("dpNotDisturbData", 33)
    DEVICE_INFO = ("dpDeviceInfo", 34)
    VOICE_PACKAGE = ("dpVoicePackage", 35)
    VOICE_LANGUAGE = ("dpVoiceLanguage", 36)
    DUST_SWITCH = ("dpDustSwitch", 37)
    CUSTOM_MODE = ("dpCustomMode", 39)
    MOP_STATE = ("dpMopState", 40)
    UNIT = ("dpUnit", 42)
    CARPET_CLEAN_PREFER = ("dpCarpetCleanPrefer", 44)
    AUTO_BOOST = ("dpAutoBoost", 45)
    CHILD_LOCK = ("dpChildLock", 47)
    DUST_SETTING = ("dpDustSetting", 50)
    MAP_SAVE_SWITCH = ("dpMapSaveSwitch", 51)
    CLEAN_RECORD = ("dpCleanRecord", 52)
    RECENT_CLEAN_RECORD = ("dpRecentCleanRecord", 53)  # NOTE: typo "dpRecendCleanRecord" in source code
    RESTRICTED_ZONE = ("dpRestrictedZone", 54)
    RESTRICTED_ZONE_UP = ("dpRestrictedZoneUp", 55)
    VIRTUAL_WALL = ("dpVirtualWall", 56)
    VIRTUAL_WALL_UP = ("dpVirtualWallUp", 57)
    ZONED = ("dpZoned", 58)
    ZONED_UP = ("dpZonedUp", 59)
    MULTI_MAP_SWITCH = ("dpMultiMapSwitch", 60)
    MULTI_MAP = ("dpMultiMap", 61)
    CUSTOMER_CLEAN = ("dpCustomerClean", 62)
    CUSTOMER_CLEAN_REQUEST = ("dpCustomerCleanRequest", 63)
    GET_CARPET = ("dpGetCarpet", 64)
    CARPET_UP = ("dpCarpetUp", 65)
    SELF_IDENTIFYING_CARPET = ("dpSelfIdentifyingCarpet", 66)
    SENSOR_LIFE = ("dpSensorLife", 67)
    RESET_SENSOR = ("dpResetSensor", 68)
    REQUEST_TIMER = ("dpRequestTimer", 69)
    REMOVE_ZONED = ("dpRemoveZoned", 70)
    REMOVE_ZONED_UP = ("dpRemoveZonedUp", 71)
    ROOM_MERGE = ("dpRoomMerge", 72)
    ROOM_SPLIT = ("dpRoomSplit", 73)
    RESET_ROOM_NAME = ("dpResetRoomName", 74)
    REQUEST_NOT_DISTURB_DATA = ("dpRequestNotDisturbData", 75)  # NOTE: typo "dpRequsetNotDisturbData" in source code
    CARPET_CLEAN_TYPE = ("dpCarpetCleanType", 76)
    BUTTON_LIGHT_SWITCH = ("dpButtonLightSwitch", 77)
    CLEAN_LINE = ("dpCleanLine", 78)
    TIME_ZONE = ("dpTimeZone", 79)
    AREA_UNIT = ("dpAreaUnit", 80)
    NET_INFO = ("dpNetInfo", 81)
    CLEAN_ORDER = ("dpCleanOrder", 82)
    ROBOT_TYPE = ("dpRobotType", 83)
    LOG_SWITCH = ("dpLogSwitch", 84)
    FLOOR_MATERIAL = ("dpFloorMaterial", 85)
    LINE_LASER_OBSTACLE_AVOIDANCE = ("dpLineLaserObstacleAvoidance", 86)
    CLEAN_PROGRESS = ("dpCleanProgress", 87)  # NOTE: typo "dpCleanProgess" in source code
    GROUND_CLEAN = ("dpGroundClean", 88)
    IGNORE_OBSTACLE = ("dpIgnoreObstacle", 89)
    FAULT = ("dpFault", 90)
    CLEAN_EXPAND = ("dpCleanExpand", 91)
    NOT_DISTURB_EXPAND = ("dpNotDisturbExpand", 92)
    TIMER_TYPE = ("dpTimerType", 93)
    CREATE_MAP_FINISHED = ("dpCreateMapFinished", 94)
    ADD_CLEAN_AREA = ("dpAddCleanArea", 95)
    ADD_CLEAN_STATE = ("dpAddCleanState", 96)
    RESTRICTED_AREA = ("dpRestrictedArea", 97)
    RESTRICTED_AREA_UP = ("dpRestrictedAreaUp", 98)
    SUSPECTED_THRESHOLD = ("dpSuspectedThreshold", 99)
    SUSPECTED_THRESHOLD_UP = ("dpSuspectedThresholdUp", 100)
    COMMON = ("dpCommon", 101)
    JUMP_SCAN = ("dpJumpScan", 101)
    REQUEST_DPS = ("dpRequestDps", 102)  # NOTE: typo "dpRequetdps" in source code
    CLIFF_RESTRICTED_AREA = ("dpCliffRestrictedArea", 102)
    CLIFF_RESTRICTED_AREA_UP = ("dpCliffRestrictedAreaUp", 103)
    BREAKPOINT_CLEAN = ("dpBreakpointClean", 104)
    VALLEY_POINT_CHARGING = ("dpValleyPointCharging", 105)
    VALLEY_POINT_CHARGING_DATA_UP = ("dpValleyPointChargingDataUp", 106)
    VALLEY_POINT_CHARGING_DATA = ("dpValleyPointChargingData", 107)
    VOICE_VERSION = ("dpVoiceVersion", 108)
    ROBOT_COUNTRY_CODE = ("dpRobotCountryCode", 109)
    HEARTBEAT = ("dpHeartbeat", 110)
    STATUS = ("dpStatus", 121)
    BATTERY = ("dpBattery", 122)
    FAN_LEVEL = ("dpFanLevel", 123)  # NOTE: typo "dpfunLevel" in source code
    WATER_LEVEL = ("dpWaterLevel", 124)
    MAIN_BRUSH_LIFE = ("dpMainBrushLife", 125)
    SIDE_BRUSH_LIFE = ("dpSideBrushLife", 126)
    FILTER_LIFE = ("dpFilterLife", 127)
    TASK_CANCEL_IN_MOTION = ("dpTaskCancelInMotion", 132)
    OFFLINE = ("dpOffline", 135)
    CLEAN_COUNT = ("dpCleanCount", 136)
    CLEAN_MODE = ("dpCleanMode", 137)
    CLEAN_TASK_TYPE = ("dpCleanTaskType", 138)
    BACK_TYPE = ("dpBackType", 139)
    CLEANING_PROGRESS = ("dpCleaningProgress", 141)
    FLEEING_GOODS = ("dpFleeingGoods", 142)
    START_CLEAN = ("dpStartClean", 201)
    START_BACK = ("dpStartBack", 202)
    START_DOCK_TASK = ("dpStartDockTask", 203)
    PAUSE = ("dpPause", 204)
    RESUME = ("dpResume", 205)
    STOP = ("dpStop", 206)
    USER_PLAN = ("dpUserPlan", 207)


class YXFanLevel(RoborockModeEnum):
    UNKNOWN = "unknown", -1
    CLOSE = "close", 0
    QUIET = "quiet", 1
    NORMAL = "normal", 2
    STRONG = "strong", 3
    MAX = "max", 4
    SUPER = "super", 8


class YXWaterLevel(RoborockModeEnum):
    UNKNOWN = "unknown", -1
    CLOSE = "close", 0
    LOW = "low", 1
    MIDDLE = "middle", 2
    HIGH = "high", 3


class YXCleanLine(RoborockModeEnum):
    FAST = "fast", 0
    DAILY = "daily", 1
    FINE = "fine", 2


class YXRoomMaterial(RoborockModeEnum):
    HORIZONTAL_FLOOR_BOARD = "horizontalfloorboard", 0
    VERTICAL_FLOOR_BOARD = "verticalfloorboard", 1
    CERAMIC_TILE = "ceramictile", 2
    OTHER = "other", 255


class YXCleanType(RoborockModeEnum):
    UNKNOWN = "unknown", -1
    BOTH_WORK = "bothwork", 1
    ONLY_SWEEP = "onlysweep", 2
    ONLY_MOP = "onlymop", 3


class YXDeviceState(RoborockModeEnum):
    UNKNOWN = "unknown", -1
    SLEEP_STATE = "sleepstate", 2
    STANDBY_STATE = "standbystate", 3
    CLEANING_STATE = "cleaningstate", 5
    TO_CHARGE_STATE = "tochargestate", 6
    REMOTEING_STATE = "remoteingstate", 7
    CHARGING_STATE = "chargingstate", 8
    PAUSE_STATE = "pausestate", 10
    FAULT_STATE = "faultstate", 12
    UPGRADE_STATE = "upgradestate", 14
    DUSTING = "dusting", 22
    CREATING_MAP_STATE = "creatingmapstate", 29
    MAP_SAVE_STATE = "mapsavestate", 99
    RE_LOCATION_STATE = "relocationstate", 101
    ROBOT_SWEEPING = "robotsweeping", 102
    ROBOT_MOPING = "robotmoping", 103
    ROBOT_SWEEP_AND_MOPING = "robotsweepandmoping", 104
    ROBOT_TRANSITIONING = "robottransitioning", 105
    ROBOT_WAIT_CHARGE = "robotwaitcharge", 108


class YXBackType(RoborockModeEnum):
    UNKNOWN = "unknown", -1
    IDLE = "idle", 0
    BACK_DUSTING = "backdusting", 4
    BACK_CHARGING = "backcharging", 5


class YXDeviceWorkMode(RoborockModeEnum):
    UNKNOWN = "unknown", -1
    BOTH_WORK = "bothwork", 1
    ONLY_SWEEP = "onlysweep", 2
    ONLY_MOP = "onlymop", 3
    CUSTOMIZED = "customized", 4
    SAVE_WORRY = "saveworry", 5
    SWEEP_MOP = "sweepmop", 6


class YXDeviceCleanTask(RoborockModeEnum):
    UNKNOWN = "unknown", -1
    IDLE = "idle", 0
    SMART = "smart", 1
    ELECTORAL = "electoral", 2
    DIVIDE_AREAS = "divideareas", 3
    CREATING_MAP = "creatingmap", 4
    PART = "part", 5


class YXDeviceDustCollectionFrequency(RoborockModeEnum):
    DAILY = "daily", 0
    INTERVAL_15 = "interval_15", 15
    INTERVAL_30 = "interval_30", 30
    INTERVAL_45 = "interval_45", 45
    INTERVAL_60 = "interval_60", 60



"""Thin wrapper around the MQTT channel for Roborock B01 Q7 devices."""

from __future__ import annotations

import asyncio
import json
import logging
from typing import Any

from roborock.devices.transport.mqtt_channel import MqttChannel
from roborock.exceptions import RoborockException
from roborock.protocols.b01_q7_protocol import (
    Q7RequestMessage,
    decode_rpc_response,
    encode_mqtt_payload,
)
from roborock.roborock_message import RoborockMessage

_LOGGER = logging.getLogger(__name__)
_TIMEOUT = 10.0


async def send_decoded_command(
    mqtt_channel: MqttChannel,
    request_message: Q7RequestMessage,
) -> dict[str, Any] | None:
    """Send a command on the MQTT channel and get a decoded response."""
    _LOGGER.debug("Sending B01 MQTT command: %s", request_message)
    roborock_message = encode_mqtt_payload(request_message)
    future: asyncio.Future[Any] = asyncio.get_running_loop().create_future()

    def find_response(response_message: RoborockMessage) -> None:
        """Handle incoming messages and resolve the future."""
        try:
            decoded_dps = decode_rpc_response(response_message)
        except RoborockException as ex:
            _LOGGER.debug(
                "Failed to decode B01 RPC response (expecting method=%s msg_id=%s): %s: %s",
                request_message.command,
                request_message.msg_id,
                response_message,
                ex,
            )
            return
        for dps_value in decoded_dps.values():
            # valid responses are JSON strings wrapped in the dps value
            if not isinstance(dps_value, str):
                _LOGGER.debug("Received unexpected response: %s", dps_value)
                continue

            try:
                inner = json.loads(dps_value)
            except (json.JSONDecodeError, TypeError):
                _LOGGER.debug("Received unexpected response: %s", dps_value)
                continue
            if isinstance(inner, dict) and inner.get("msgId") == str(request_message.msg_id):
                _LOGGER.debug("Received query response: %s", inner)
                # Check for error code (0 = success, non-zero = error)
                code = inner.get("code", 0)
                if code != 0:
                    error_msg = f"B01 command failed with code {code} ({request_message})"
                    _LOGGER.debug("B01 error response: %s", error_msg)
                    if not future.done():
                        future.set_exception(RoborockException(error_msg))
                    return
                data = inner.get("data")
                # All get commands should be dicts
                if request_message.command.endswith(".get") and not isinstance(data, dict):
                    if not future.done():
                        future.set_exception(
                            RoborockException(f"Unexpected data type for response {data} ({request_message})")
                        )
                    return
                if not future.done():
                    future.set_result(data)

    unsub = await mqtt_channel.subscribe(find_response)

    _LOGGER.debug("Sending MQTT message: %s", roborock_message)
    try:
        await mqtt_channel.publish(roborock_message)
        return await asyncio.wait_for(future, timeout=_TIMEOUT)
    except TimeoutError as ex:
        raise RoborockException(f"B01 command timed out after {_TIMEOUT}s ({request_message})") from ex
    except RoborockException as ex:
        _LOGGER.warning(
            "Error sending B01 decoded command (%ss): %s",
            request_message,
            ex,
        )
        raise

    except Exception as ex:
        _LOGGER.exception(
            "Error sending B01 decoded command (%ss): %s",
            request_message,
            ex,
        )
        raise
    finally:
        unsub()



from __future__ import annotations

from dataclasses import dataclass, field
from enum import StrEnum

from roborock import RoborockEnum
from roborock.util import get_next_int, get_timestamp


class RoborockMessageProtocol(RoborockEnum):
    HELLO_REQUEST = 0
    HELLO_RESPONSE = 1
    PING_REQUEST = 2
    PING_RESPONSE = 3
    GENERAL_REQUEST = 4
    GENERAL_RESPONSE = 5
    RPC_REQUEST = 101
    RPC_RESPONSE = 102
    MAP_RESPONSE = 301


class RoborockDataProtocol(RoborockEnum):
    ERROR_CODE = 120
    STATE = 121
    BATTERY = 122
    FAN_POWER = 123
    WATER_BOX_MODE = 124
    MAIN_BRUSH_WORK_TIME = 125
    SIDE_BRUSH_WORK_TIME = 126
    FILTER_WORK_TIME = 127
    ADDITIONAL_PROPS = 128
    TASK_COMPLETE = 130
    TASK_CANCEL_LOW_POWER = 131
    TASK_CANCEL_IN_MOTION = 132
    CHARGE_STATUS = 133
    DRYING_STATUS = 134
    OFFLINE_STATUS = 135

    @classmethod
    def _missing_(cls: type[RoborockEnum], key) -> RoborockEnum:
        raise ValueError("%s not a valid key for Data Protocol", key)


class RoborockDyadDataProtocol(RoborockEnum):
    DRYING_STATUS = 134
    START = 200
    STATUS = 201
    SELF_CLEAN_MODE = 202
    SELF_CLEAN_LEVEL = 203
    WARM_LEVEL = 204
    CLEAN_MODE = 205
    SUCTION = 206
    WATER_LEVEL = 207
    BRUSH_SPEED = 208
    POWER = 209
    COUNTDOWN_TIME = 210
    AUTO_SELF_CLEAN_SET = 212
    AUTO_DRY = 213
    MESH_LEFT = 214
    BRUSH_LEFT = 215
    ERROR = 216
    MESH_RESET = 218
    BRUSH_RESET = 219
    VOLUME_SET = 221
    STAND_LOCK_AUTO_RUN = 222
    AUTO_SELF_CLEAN_SET_MODE = 223
    AUTO_DRY_MODE = 224
    SILENT_DRY_DURATION = 225
    SILENT_MODE = 226
    SILENT_MODE_START_TIME = 227
    SILENT_MODE_END_TIME = 228
    RECENT_RUN_TIME = 229
    TOTAL_RUN_TIME = 230
    FEATURE_INFO = 235
    RECOVER_SETTINGS = 236
    DRY_COUNTDOWN = 237
    ID_QUERY = 10000
    F_C = 10001
    SCHEDULE_TASK = 10002
    SND_SWITCH = 10003
    SND_STATE = 10004
    PRODUCT_INFO = 10005
    PRIVACY_INFO = 10006
    OTA_NFO = 10007
    RPC_REQUEST = 10101
    RPC_RESPONSE = 10102


class RoborockZeoProtocol(RoborockEnum):
    START = 200  # rw
    PAUSE = 201  # rw
    SHUTDOWN = 202  # rw
    STATE = 203  # ro
    MODE = 204  # rw
    PROGRAM = 205  # rw
    CHILD_LOCK = 206  # rw
    TEMP = 207  # rw
    RINSE_TIMES = 208  # rw
    SPIN_LEVEL = 209  # rw
    DRYING_MODE = 210  # rw
    DETERGENT_SET = 211  # rw
    SOFTENER_SET = 212  # rw
    DETERGENT_TYPE = 213  # rw
    SOFTENER_TYPE = 214  # rw
    COUNTDOWN = 217  # rw
    WASHING_LEFT = 218  # ro
    DOORLOCK_STATE = 219  # ro
    ERROR = 220  # ro
    CUSTOM_PARAM_SAVE = 221  # rw
    CUSTOM_PARAM_GET = 222  # ro
    SOUND_SET = 223  # rw
    TIMES_AFTER_CLEAN = 224  # ro
    DEFAULT_SETTING = 225  # rw
    DETERGENT_EMPTY = 226  # ro
    SOFTENER_EMPTY = 227  # ro
    LIGHT_SETTING = 229  # rw
    DETERGENT_VOLUME = 230  # rw
    SOFTENER_VOLUME = 231  # rw
    APP_AUTHORIZATION = 232  # rw
    ID_QUERY = 10000
    F_C = 10001
    SND_STATE = 10004
    PRODUCT_INFO = 10005
    PRIVACY_INFO = 10006
    OTA_NFO = 10007
    WASHING_LOG = 10008
    RPC_REQ = 10101
    RPC_RESp = 10102


class RoborockB01Protocol(RoborockEnum):
    RPC_REQUEST = 101
    RPC_RESPONSE = 102
    ERROR_CODE = 120
    STATE = 121
    BATTERY = 122
    FAN_POWER = 123
    WATER_BOX_MODE = 124
    MAIN_BRUSH_LIFE = 125
    SIDE_BRUSH_LIFE = 126
    FILTER_LIFE = 127
    OFFLINE_STATUS = 135
    CLEAN_TIMES = 136
    CLEANING_PREFERENCE = 137
    CLEAN_TASK_TYPE = 138
    BACK_TYPE = 139
    DOCK_TASK_TYPE = 140
    CLEANING_PROGRESS = 141
    FC_STATE = 142
    START_CLEAN_TASK = 201
    START_BACK_DOCK_TASK = 202
    START_DOCK_TASK = 203
    PAUSE = 204
    RESUME = 205
    STOP = 206
    CEIP = 207


class RoborockB01Props(StrEnum):
    """Properties requested by the Roborock B01 model."""

    STATUS = "status"
    FAULT = "fault"
    WIND = "wind"
    WATER = "water"
    MODE = "mode"
    QUANTITY = "quantity"
    ALARM = "alarm"
    VOLUME = "volume"
    HYPA = "hypa"
    MAIN_BRUSH = "main_brush"
    SIDE_BRUSH = "side_brush"
    MOP_LIFE = "mop_life"
    MAIN_SENSOR = "main_sensor"
    NET_STATUS = "net_status"
    REPEAT_STATE = "repeat_state"
    TANK_STATE = "tank_state"
    SWEEP_TYPE = "sweep_type"
    CLEAN_PATH_PREFERENCE = "clean_path_preference"
    CLOTH_STATE = "cloth_state"
    TIME_ZONE = "time_zone"
    TIME_ZONE_INFO = "time_zone_info"
    LANGUAGE = "language"
    CLEANING_TIME = "cleaning_time"
    REAL_CLEAN_TIME = "real_clean_time"
    CLEANING_AREA = "cleaning_area"
    CUSTOM_TYPE = "custom_type"
    SOUND = "sound"
    WORK_MODE = "work_mode"
    STATION_ACT = "station_act"
    CHARGE_STATE = "charge_state"
    CURRENT_MAP_ID = "current_map_id"
    MAP_NUM = "map_num"
    DUST_ACTION = "dust_action"
    QUIET_IS_OPEN = "quiet_is_open"
    QUIET_BEGIN_TIME = "quiet_begin_time"
    QUIET_END_TIME = "quiet_end_time"
    CLEAN_FINISH = "clean_finish"
    VOICE_TYPE = "voice_type"
    VOICE_TYPE_VERSION = "voice_type_version"
    ORDER_TOTAL = "order_total"
    BUILD_MAP = "build_map"
    PRIVACY = "privacy"
    DUST_AUTO_STATE = "dust_auto_state"
    DUST_FREQUENCY = "dust_frequency"
    CHILD_LOCK = "child_lock"
    MULTI_FLOOR = "multi_floor"
    MAP_SAVE = "map_save"
    LIGHT_MODE = "light_mode"
    GREEN_LASER = "green_laser"
    DUST_BAG_USED = "dust_bag_used"
    ORDER_SAVE_MODE = "order_save_mode"
    MANUFACTURER = "manufacturer"
    BACK_TO_WASH = "back_to_wash"
    CHARGE_STATION_TYPE = "charge_station_type"
    PV_CUT_CHARGE = "pv_cut_charge"
    PV_CHARGING = "pv_charging"
    SERIAL_NUMBER = "serial_number"
    RECOMMEND = "recommend"
    ADD_SWEEP_STATUS = "add_sweep_status"


ROBOROCK_DATA_STATUS_PROTOCOL = [
    RoborockDataProtocol.ERROR_CODE,
    RoborockDataProtocol.STATE,
    RoborockDataProtocol.BATTERY,
    RoborockDataProtocol.FAN_POWER,
    RoborockDataProtocol.WATER_BOX_MODE,
    RoborockDataProtocol.CHARGE_STATUS,
]

ROBOROCK_DATA_CONSUMABLE_PROTOCOL = [
    RoborockDataProtocol.MAIN_BRUSH_WORK_TIME,
    RoborockDataProtocol.SIDE_BRUSH_WORK_TIME,
    RoborockDataProtocol.FILTER_WORK_TIME,
]


@dataclass
class RoborockMessage:
    protocol: RoborockMessageProtocol
    payload: bytes | None = None
    seq: int = field(default_factory=lambda: get_next_int(100000, 999999))
    version: bytes = b"1.0"
    random: int = field(default_factory=lambda: get_next_int(10000, 99999))
    timestamp: int = field(default_factory=lambda: get_timestamp())



"""Status trait for Q10 B01 devices."""

import logging
from typing import Any

from roborock.data.b01_q10.b01_q10_code_mappings import B01_Q10_DP
from roborock.data.b01_q10.b01_q10_containers import Q10Status

from .common import DpsDataConverter, TraitUpdateListener

_LOGGER = logging.getLogger(__name__)

_CONVERTER = DpsDataConverter.from_dataclass(Q10Status)


class StatusTrait(Q10Status, TraitUpdateListener):
    """Trait for managing the status of Q10 Roborock devices.

    This is a thin wrapper around Q10Status that provides the Trait interface.
    The current values reflect the most recently received data from the device.
    New values can be requested through the `Q10PropertiesApi`'s `refresh` method.
    """

    def __init__(self) -> None:
        """Initialize the status trait."""
        super().__init__()
        TraitUpdateListener.__init__(self, logger=_LOGGER)

    def update_from_dps(self, decoded_dps: dict[B01_Q10_DP, Any]) -> None:
        """Update the trait from raw DPS data."""
        if _CONVERTER.update_from_dps(self, decoded_dps):
            self._notify_update()



"""Module for parsing v1 Roborock map content."""

import io
import logging
from dataclasses import dataclass, field

from vacuum_map_parser_base.config.color import ColorsPalette, SupportedColor
from vacuum_map_parser_base.config.drawable import Drawable
from vacuum_map_parser_base.config.image_config import ImageConfig
from vacuum_map_parser_base.config.size import Size, Sizes
from vacuum_map_parser_base.map_data import MapData
from vacuum_map_parser_roborock.map_data_parser import RoborockMapDataParser

from roborock.exceptions import RoborockException

_LOGGER = logging.getLogger(__name__)

DEFAULT_DRAWABLES = {
    Drawable.CHARGER: True,
    Drawable.CLEANED_AREA: False,
    Drawable.GOTO_PATH: False,
    Drawable.IGNORED_OBSTACLES: False,
    Drawable.IGNORED_OBSTACLES_WITH_PHOTO: False,
    Drawable.MOP_PATH: False,
    Drawable.NO_CARPET_AREAS: False,
    Drawable.NO_GO_AREAS: False,
    Drawable.NO_MOPPING_AREAS: False,
    Drawable.OBSTACLES: False,
    Drawable.OBSTACLES_WITH_PHOTO: False,
    Drawable.PATH: True,
    Drawable.PREDICTED_PATH: False,
    Drawable.VACUUM_POSITION: True,
    Drawable.VIRTUAL_WALLS: False,
    Drawable.ZONES: False,
}
DEFAULT_MAP_SCALE = 4
MAP_FILE_FORMAT = "PNG"


def _default_drawable_factory() -> list[Drawable]:
    return [drawable for drawable, default_value in DEFAULT_DRAWABLES.items() if default_value]


@dataclass
class MapParserConfig:
    """Configuration for the Roborock map parser."""

    drawables: list[Drawable] = field(default_factory=_default_drawable_factory)
    """List of drawables to include in the map rendering."""

    show_background: bool = True
    """Whether to show the background of the map."""

    show_walls: bool = True
    """Whether to show the walls of the map."""

    show_rooms: bool = True
    """Whether to show the rooms of the map."""

    map_scale: int = DEFAULT_MAP_SCALE
    """Scale factor for the map."""


@dataclass
class ParsedMapData:
    """Roborock Map Data.

    This class holds the parsed map data and the rendered image.
    """

    image_content: bytes | None
    """The rendered image of the map in PNG format."""

    map_data: MapData | None
    """The parsed map data which contains metadata for points on the map."""


class MapParser:
    """Roborock Map Parser.

    This class is used to parse the map data from the device and render it into an image.
    """

    def __init__(self, config: MapParserConfig) -> None:
        """Initialize the MapParser."""
        self._map_parser = _create_map_data_parser(config)

    def parse(self, map_bytes: bytes) -> ParsedMapData | None:
        """Parse map_bytes and return MapData and the image."""
        try:
            parsed_map = self._map_parser.parse(map_bytes)
        except (IndexError, ValueError) as err:
            raise RoborockException("Failed to parse map data") from err
        if parsed_map.image is None:
            raise RoborockException("Failed to render map image")
        img_byte_arr = io.BytesIO()
        parsed_map.image.data.save(img_byte_arr, format=MAP_FILE_FORMAT)
        return ParsedMapData(image_content=img_byte_arr.getvalue(), map_data=parsed_map)


def _create_map_data_parser(config: MapParserConfig) -> RoborockMapDataParser:
    """Create a RoborockMapDataParser based on the config entry."""
    color_dicts = {}
    room_colors = {}

    if not config.show_background:
        color_dicts[SupportedColor.MAP_OUTSIDE] = (0, 0, 0, 0)

    if not config.show_walls:
        color_dicts[SupportedColor.GREY_WALL] = (0, 0, 0, 0)
        color_dicts[SupportedColor.MAP_WALL] = (0, 0, 0, 0)
        color_dicts[SupportedColor.MAP_WALL_V2] = (0, 0, 0, 0)

    if not config.show_rooms:
        room_colors = {str(x): (0, 0, 0, 0) for x in range(1, 32)}

    return RoborockMapDataParser(
        ColorsPalette(color_dicts, room_colors),
        Sizes({k: v * config.map_scale for k, v in Sizes.SIZES.items() if k != Size.MOP_PATH_WIDTH}),
        config.drawables,
        ImageConfig(scale=config.map_scale),
        [],
    )



from ..code_mappings import RoborockEnum


class RoborockFinishReason(RoborockEnum):
    manual_interrupt = 21  # Cleaning interrupted by user
    cleanup_interrupted = 24  # Cleanup interrupted
    manual_interrupt_2 = 21
    manual_interrupt_12 = 29
    breakpoint = 32  # Could not continue cleaning
    breakpoint_2 = 33
    cleanup_interrupted_2 = 34
    manual_interrupt_3 = 35
    manual_interrupt_4 = 36
    manual_interrupt_5 = 37
    manual_interrupt_6 = 43
    locate_fail = 45  # Positioning Failed
    cleanup_interrupted_3 = 64
    locate_fail_2 = 65
    manual_interrupt_7 = 48
    manual_interrupt_8 = 49
    manual_interrupt_9 = 50
    cleanup_interrupted_4 = 51
    finished_cleaning = 52  # Finished cleaning
    finished_cleaning_2 = 54
    finished_cleaning_3 = 55
    finished_cleaning_4 = 56
    finished_clenaing_5 = 57
    manual_interrupt_10 = 60
    area_unreachable = 61  # Area unreachable
    area_unreachable_2 = 62
    washing_error = 67  # Washing error
    back_to_wash_failure = 68  # Failed to return to the dock
    cleanup_interrupted_5 = 101
    breakpoint_4 = 102
    manual_interrupt_11 = 103
    cleanup_interrupted_6 = 104
    cleanup_interrupted_7 = 105
    cleanup_interrupted_8 = 106
    cleanup_interrupted_9 = 107
    cleanup_interrupted_10 = 109
    cleanup_interrupted_11 = 110
    patrol_success = 114  # Cruise completed
    patrol_fail = 115  # Cruise failed
    pet_patrol_success = 116  # Pet found
    pet_patrol_fail = 117  # Pet found failed


class RoborockInCleaning(RoborockEnum):
    complete = 0
    global_clean_not_complete = 1
    zone_clean_not_complete = 2
    segment_clean_not_complete = 3


class RoborockCleanType(RoborockEnum):
    all_zone = 1
    draw_zone = 2
    select_zone = 3
    quick_build = 4
    video_patrol = 5
    pet_patrol = 6


class RoborockStartType(RoborockEnum):
    button = 1
    app = 2
    schedule = 3
    mi_home = 4
    quick_start = 5
    voice_control = 13
    routines = 101
    alexa = 801
    google = 802
    ifttt = 803
    yandex = 804
    homekit = 805
    xiaoai = 806
    tmall_genie = 807
    duer = 808
    dingdong = 809
    siri = 810
    clova = 811
    wechat = 901
    alipay = 902
    aqara = 903
    hisense = 904
    huawei = 905
    widget_launch = 820
    smart_watch = 821


class RoborockDssCodes(RoborockEnum):
    @classmethod
    def _missing_(cls: type[RoborockEnum], key) -> RoborockEnum:
        # If the calculated value is not provided, then it should be viewed as okay.
        # As the math will sometimes result in you getting numbers that don't matter.
        return cls.okay  # type: ignore


class ClearWaterBoxStatus(RoborockDssCodes):
    """Status of the clear water box."""

    okay = 0
    out_of_water = 1
    out_of_water_2 = 38
    refill_error = 48


class DirtyWaterBoxStatus(RoborockDssCodes):
    """Status of the dirty water box."""

    okay = 0
    full_not_installed = 1
    full_not_installed_2 = 39
    drain_error = 49


class DustBagStatus(RoborockDssCodes):
    """Status of the dust bag."""

    okay = 0
    not_installed = 1
    full = 34


class CleanFluidStatus(RoborockDssCodes):
    """Status of the cleaning fluid container."""

    empty_not_installed = 1
    okay = 2


class RoborockErrorCode(RoborockEnum):
    none = 0
    lidar_blocked = 1
    bumper_stuck = 2
    wheels_suspended = 3
    cliff_sensor_error = 4
    main_brush_jammed = 5
    side_brush_jammed = 6
    wheels_jammed = 7
    robot_trapped = 8
    no_dustbin = 9
    strainer_error = 10  # Filter is wet or blocked
    compass_error = 11  # Strong magnetic field detected
    low_battery = 12
    charging_error = 13
    battery_error = 14
    wall_sensor_dirty = 15
    robot_tilted = 16
    side_brush_error = 17
    fan_error = 18
    dock = 19  # Dock not connected to power
    optical_flow_sensor_dirt = 20
    vertical_bumper_pressed = 21
    dock_locator_error = 22
    return_to_dock_fail = 23
    nogo_zone_detected = 24
    visual_sensor = 25  # Camera error
    light_touch = 26  # Wall sensor error
    vibrarise_jammed = 27
    robot_on_carpet = 28
    filter_blocked = 29
    invisible_wall_detected = 30
    cannot_cross_carpet = 31
    internal_error = 32
    collect_dust_error_3 = 34  # Clean auto-empty dock
    collect_dust_error_4 = 35  # Auto empty dock voltage error
    mopping_roller_1 = 36  # Wash roller may be jammed
    mopping_roller_error_2 = 37  # wash roller not lowered properly
    clear_water_box_hoare = 38  # Check the clean water tank
    dirty_water_box_hoare = 39  # Check the dirty water tank
    sink_strainer_hoare = 40  # Reinstall the water filter
    clear_water_box_exception = 41  # Clean water tank empty
    clear_brush_exception = 42  # Check that the water filter has been correctly installed
    clear_brush_exception_2 = 43  # Positioning button error
    filter_screen_exception = 44  # Clean the dock water filter
    mopping_roller_2 = 45  # Wash roller may be jammed
    up_water_exception = 48
    drain_water_exception = 49
    temperature_protection = 51  # Unit temperature protection
    clean_carousel_exception = 52
    clean_carousel_water_full = 53
    water_carriage_drop = 54
    check_clean_carouse = 55
    audio_error = 56


class RoborockFanPowerCode(RoborockEnum):
    """Describes the fan power of the vacuum cleaner."""

    # Fan speeds should have the first letter capitalized - as there is no way to change the name in translations as
    # far as I am aware


class RoborockFanSpeedV1(RoborockFanPowerCode):
    silent = 38
    standard = 60
    medium = 77
    turbo = 90


class RoborockFanSpeedV2(RoborockFanPowerCode):
    silent = 101
    balanced = 102
    turbo = 103
    max = 104
    gentle = 105
    auto = 106


class RoborockFanSpeedV3(RoborockFanPowerCode):
    silent = 38
    standard = 60
    medium = 75
    turbo = 100


class RoborockFanSpeedE2(RoborockFanPowerCode):
    gentle = 41
    silent = 50
    standard = 68
    medium = 79
    turbo = 100


class RoborockFanSpeedS7(RoborockFanPowerCode):
    off = 105
    quiet = 101
    balanced = 102
    turbo = 103
    max = 104
    custom = 106


class RoborockFanSpeedS7MaxV(RoborockFanPowerCode):
    off = 105
    quiet = 101
    balanced = 102
    turbo = 103
    max = 104
    custom = 106
    max_plus = 108


class RoborockFanSpeedS6Pure(RoborockFanPowerCode):
    gentle = 105
    quiet = 101
    balanced = 102
    turbo = 103
    max = 104
    custom = 106


class RoborockFanSpeedQ7Max(RoborockFanPowerCode):
    quiet = 101
    balanced = 102
    turbo = 103
    max = 104


class RoborockFanSpeedQRevoMaster(RoborockFanPowerCode):
    off = 105
    quiet = 101
    balanced = 102
    turbo = 103
    max = 104
    custom = 106
    max_plus = 108
    smart_mode = 110


class RoborockFanSpeedQRevoCurv(RoborockFanPowerCode):
    quiet = 101
    balanced = 102
    turbo = 103
    max = 104
    off = 105
    custom = 106
    max_plus = 108
    smart_mode = 110


class RoborockFanSpeedQRevoMaxV(RoborockFanPowerCode):
    off = 105
    quiet = 101
    balanced = 102
    turbo = 103
    max = 104
    custom = 106
    max_plus = 108
    smart_mode = 110


class RoborockFanSpeedP10(RoborockFanPowerCode):
    off = 105
    quiet = 101
    balanced = 102
    turbo = 103
    max = 104
    custom = 106
    max_plus = 108
    smart_mode = 110


class RoborockFanSpeedS8MaxVUltra(RoborockFanPowerCode):
    off = 105
    quiet = 101
    balanced = 102
    turbo = 103
    max = 104
    custom = 106
    max_plus = 108
    smart_mode = 110


class RoborockFanSpeedSaros10(RoborockFanPowerCode):
    off = 105
    quiet = 101
    balanced = 102
    turbo = 103
    max = 104
    custom = 106
    max_plus = 108
    smart_mode = 110


class RoborockFanSpeedSaros10R(RoborockFanPowerCode):
    off = 105
    quiet = 101
    balanced = 102
    turbo = 103
    max = 104
    custom = 106
    max_plus = 108
    smart_mode = 110


class RoborockMopModeCode(RoborockEnum):
    """Describes the mop mode of the vacuum cleaner."""


class RoborockMopModeQRevoCurv(RoborockMopModeCode):
    standard = 300
    deep = 301
    custom = 302
    deep_plus = 303
    fast = 304
    smart_mode = 306


class RoborockMopModeS7(RoborockMopModeCode):
    """Describes the mop mode of the vacuum cleaner."""

    standard = 300
    deep = 301
    custom = 302
    deep_plus = 303


class RoborockMopModeS8ProUltra(RoborockMopModeCode):
    standard = 300
    deep = 301
    deep_plus = 303
    fast = 304
    custom = 302
    smart_mode = 306


class RoborockMopModeS8MaxVUltra(RoborockMopModeCode):
    standard = 300
    deep = 301
    custom = 302
    deep_plus = 303
    fast = 304
    deep_plus_pearl = 305
    smart_mode = 306


class RoborockMopModeSaros10R(RoborockMopModeCode):
    standard = 300
    deep = 301
    custom = 302
    deep_plus = 303
    fast = 304
    smart_mode = 306


class RoborockMopModeQRevoMaster(RoborockMopModeCode):
    standard = 300
    deep = 301
    custom = 302
    deep_plus = 303
    fast = 304
    smart_mode = 306


class RoborockMopModeQRevoMaxV(RoborockMopModeCode):
    standard = 300
    deep = 301
    custom = 302
    deep_plus = 303
    fast = 304
    smart_mode = 306


class RoborockMopModeSaros10(RoborockMopModeCode):
    standard = 300
    deep = 301
    custom = 302
    deep_plus = 303
    fast = 304
    smart_mode = 306


class RoborockMopIntensityCode(RoborockEnum):
    """Describes the mop intensity of the vacuum cleaner."""


class RoborockMopIntensityS7(RoborockMopIntensityCode):
    """Describes the mop intensity of the vacuum cleaner."""

    off = 200
    mild = 201
    moderate = 202
    intense = 203
    custom = 204


class RoborockMopIntensityV2(RoborockMopIntensityCode):
    """Describes the mop intensity of the vacuum cleaner."""

    off = 200
    low = 201
    medium = 202
    high = 203
    custom = 207


class RoborockMopIntensityQRevoMaster(RoborockMopIntensityCode):
    """Describes the mop intensity of the vacuum cleaner."""

    off = 200
    low = 201
    medium = 202
    high = 203
    custom = 204
    custom_water_flow = 207
    smart_mode = 209


class RoborockMopIntensityQRevoCurv(RoborockMopIntensityCode):
    off = 200
    low = 201
    medium = 202
    high = 203
    custom = 204
    custom_water_flow = 207
    smart_mode = 209


class RoborockMopIntensityQRevoMaxV(RoborockMopIntensityCode):
    off = 200
    low = 201
    medium = 202
    high = 203
    custom = 204
    custom_water_flow = 207
    smart_mode = 209


class RoborockMopIntensityP10(RoborockMopIntensityCode):
    """Describes the mop intensity of the vacuum cleaner."""

    off = 200
    low = 201
    medium = 202
    high = 203
    custom = 204
    custom_water_flow = 207
    smart_mode = 209


class RoborockMopIntensityS8MaxVUltra(RoborockMopIntensityCode):
    off = 200
    low = 201
    medium = 202
    high = 203
    custom = 204
    max = 208
    smart_mode = 209
    custom_water_flow = 207


class RoborockMopIntensitySaros10(RoborockMopIntensityCode):
    off = 200
    mild = 201
    standard = 202
    intense = 203
    extreme = 208
    custom = 204
    smart_mode = 209


class RoborockMopIntensitySaros10R(RoborockMopIntensityCode):
    off = 200
    low = 201
    medium = 202
    high = 203
    custom = 204
    extreme = 250
    vac_followed_by_mop = 235
    smart_mode = 209


class RoborockMopIntensityS5Max(RoborockMopIntensityCode):
    """Describes the mop intensity of the vacuum cleaner."""

    off = 200
    low = 201
    medium = 202
    high = 203
    custom = 204
    custom_water_flow = 207


class RoborockMopIntensityS6MaxV(RoborockMopIntensityCode):
    """Describes the mop intensity of the vacuum cleaner."""

    off = 200
    low = 201
    medium = 202
    high = 203
    custom = 204
    custom_water_flow = 207


class RoborockMopIntensityQ7Max(RoborockMopIntensityCode):
    """Describes the mop intensity of the vacuum cleaner."""

    off = 200
    low = 201
    medium = 202
    high = 203
    custom_water_flow = 207


class RoborockDockErrorCode(RoborockEnum):
    """Describes the error code of the dock."""

    ok = 0
    duct_blockage = 34
    water_empty = 38
    waste_water_tank_full = 39
    maintenance_brush_jammed = 42
    dirty_tank_latch_open = 44
    no_dustbin = 46
    cleaning_tank_full_or_blocked = 53


class RoborockDockTypeCode(RoborockEnum):
    unknown = -9999
    no_dock = 0
    auto_empty_dock = 1
    empty_wash_fill_dock = 3
    auto_empty_dock_pure = 5
    s7_max_ultra_dock = 6
    s8_dock = 7
    p10_dock = 8
    p10_pro_dock = 9
    s8_maxv_ultra_dock = 10
    qrevo_master_dock = 14
    qrevo_s_dock = 15
    saros_r10_dock = 16
    qrevo_curv_dock = 17
    saros_10_dock = 18


class RoborockDockDustCollectionModeCode(RoborockEnum):
    """Describes the dust collection mode of the vacuum cleaner."""

    # TODO: Get the correct values for various different docks
    unknown = -9999
    smart = 0
    light = 1
    balanced = 2
    max = 4


class RoborockDockWashTowelModeCode(RoborockEnum):
    """Describes the wash towel mode of the vacuum cleaner."""

    # TODO: Get the correct values for various different docks
    unknown = -9999
    light = 0
    balanced = 1
    deep = 2
    smart = 10


class RoborockStateCode(RoborockEnum):
    unknown = 0
    starting = 1
    charger_disconnected = 2
    idle = 3
    remote_control_active = 4
    cleaning = 5
    returning_home = 6
    manual_mode = 7
    charging = 8
    charging_problem = 9
    paused = 10
    spot_cleaning = 11
    error = 12
    shutting_down = 13
    updating = 14
    docking = 15
    going_to_target = 16
    zoned_cleaning = 17
    segment_cleaning = 18
    emptying_the_bin = 22  # on s7+
    washing_the_mop = 23  # on a46
    washing_the_mop_2 = 25
    going_to_wash_the_mop = 26  # on a46
    in_call = 28
    mapping = 29
    egg_attack = 30
    patrol = 32
    attaching_the_mop = 33  # on g20s ultra
    detaching_the_mop = 34  # on g20s ultra
    charging_complete = 100
    device_offline = 101
    locked = 103
    air_drying_stopping = 202
    robot_status_mopping = 6301
    clean_mop_cleaning = 6302
    clean_mop_mopping = 6303
    segment_mopping = 6304
    segment_clean_mop_cleaning = 6305
    segment_clean_mop_mopping = 6306
    zoned_mopping = 6307
    zoned_clean_mop_cleaning = 6308
    zoned_clean_mop_mopping = 6309
    back_to_dock_washing_duster = 6310



"""Clean summary / clean records trait for B01 Q7 devices.

For B01/Q7, the Roborock app uses `service.get_record_list` which returns totals
and a `record_list` whose items contain a JSON string in `detail`.
"""

from __future__ import annotations

import logging

from roborock import CleanRecordDetail, CleanRecordList, CleanRecordSummary
from roborock.devices.rpc.b01_q7_channel import send_decoded_command
from roborock.devices.traits import Trait
from roborock.devices.transport.mqtt_channel import MqttChannel
from roborock.exceptions import RoborockException
from roborock.protocols.b01_q7_protocol import Q7RequestMessage
from roborock.roborock_typing import RoborockB01Q7Methods

__all__ = [
    "CleanSummaryTrait",
]

_LOGGER = logging.getLogger(__name__)


class CleanSummaryTrait(CleanRecordSummary, Trait):
    """B01/Q7 clean summary + clean record access (via record list service)."""

    def __init__(self, channel: MqttChannel) -> None:
        """Initialize the clean summary trait.

        Args:
            channel: MQTT channel used to communicate with the device.
        """
        super().__init__()
        self._channel = channel

    async def refresh(self) -> None:
        """Refresh totals and last record detail from the device."""
        record_list = await self._get_record_list()

        self.total_time = record_list.total_time
        self.total_area = record_list.total_area
        self.total_count = record_list.total_count

        details = await self._get_clean_record_details(record_list=record_list)
        self.last_record_detail = details[0] if details else None

    async def _get_record_list(self) -> CleanRecordList:
        """Fetch the raw device clean record list (`service.get_record_list`)."""
        result = await send_decoded_command(
            self._channel,
            Q7RequestMessage(dps=10000, command=RoborockB01Q7Methods.GET_RECORD_LIST, params={}),
        )

        if not isinstance(result, dict):
            raise TypeError(f"Unexpected response type for GET_RECORD_LIST: {type(result).__name__}: {result!r}")
        return CleanRecordList.from_dict(result)

    async def _get_clean_record_details(self, *, record_list: CleanRecordList) -> list[CleanRecordDetail]:
        """Return parsed record detail objects (newest-first)."""
        details: list[CleanRecordDetail] = []
        for item in record_list.record_list:
            try:
                parsed = item.detail_parsed
            except RoborockException as ex:
                # Rather than failing if something goes wrong here, we should fail and log to tell the user.
                _LOGGER.debug("Failed to parse record detail: %s", ex)
                continue
            if parsed is not None:
                details.append(parsed)

        # The server returns the newest record at the end of record_list; reverse so newest is first (index 0).
        details.reverse()
        return details



"""This module is meant to hold dataclasses and codemappings for various devices and protocols."""

from .b01_q7 import *
from .b01_q10 import *
from .code_mappings import *
from .containers import *
from .dyad import *
from .v1 import *
from .zeo import *



"""Trait for managing room mappings on Roborock devices."""

import logging
from dataclasses import dataclass

from roborock.data import HomeData, HomeDataRoom, NamedRoomMapping, RoborockBase
from roborock.devices.traits.v1 import common
from roborock.roborock_typing import RoborockCommand
from roborock.web_api import UserWebApiClient

_LOGGER = logging.getLogger(__name__)


@dataclass
class Rooms(RoborockBase):
    """Dataclass representing a collection of room mappings."""

    rooms: list[NamedRoomMapping] | None = None
    """List of room mappings."""

    @property
    def room_map(self) -> dict[int, NamedRoomMapping]:
        """Returns a mapping of segment_id to NamedRoomMapping."""
        if self.rooms is None:
            return {}
        return {room.segment_id: room for room in self.rooms}


class RoomsTrait(Rooms, common.V1TraitMixin):
    """Trait for managing the room mappings of Roborock devices."""

    command = RoborockCommand.GET_ROOM_MAPPING

    def __init__(self, home_data: HomeData, web_api: UserWebApiClient) -> None:
        """Initialize the RoomsTrait."""
        super().__init__()
        self._home_data = home_data
        self._web_api = web_api
        self._discovered_iot_ids: set[str] = set()

    async def refresh(self) -> None:
        """Refresh room mappings and backfill unknown room names from the web API."""
        response = await self.rpc_channel.send_command(self.command)
        if not isinstance(response, list):
            raise ValueError(f"Unexpected RoomsTrait response format: {response!r}")

        segment_map = _extract_segment_map(response)
        # Track all iot ids seen before. Refresh the room list when new ids are found.
        new_iot_ids = set(segment_map.values()) - set(self._home_data.rooms_map.keys())
        if new_iot_ids - self._discovered_iot_ids:
            _LOGGER.debug("Refreshing room list to discover new room names")
            if updated_rooms := await self._refresh_rooms():
                _LOGGER.debug("Updating rooms: %s", list(updated_rooms))
                self._home_data.rooms = updated_rooms
            self._discovered_iot_ids.update(new_iot_ids)

        new_data = self._parse_rooms(segment_map, self._home_data.rooms_name_map)
        self._update_trait_values(new_data)
        _LOGGER.debug("Refreshed %s: %s", self.__class__.__name__, new_data)

    @staticmethod
    def _parse_rooms(
        segment_map: dict[int, str],
        name_map: dict[str, str],
    ) -> Rooms:
        """Parse the response from the device into a list of NamedRoomMapping."""
        return Rooms(
            rooms=[
                NamedRoomMapping(segment_id=segment_id, iot_id=iot_id, raw_name=name_map.get(iot_id))
                for segment_id, iot_id in segment_map.items()
            ]
        )

    async def _refresh_rooms(self) -> list[HomeDataRoom]:
        """Fetch the latest rooms from the web API."""
        try:
            return await self._web_api.get_rooms()
        except Exception:
            _LOGGER.debug("Failed to fetch rooms from web API", exc_info=True)
            return []


def _extract_segment_map(response: list) -> dict[int, str]:
    """Extract a segment_id -> iot_id mapping from the response.

    The response format can be either a flat list of [segment_id, iot_id] or a
    list of lists, where each inner list is a pair of [segment_id, iot_id]. This
    function normalizes the response into a dict of segment_id to iot_id.

    NOTE: We currently only partial samples of the room mapping formats, so
    improving test coverage with samples from a real device with this format
    would be helpful.
    """
    if len(response) == 2 and not isinstance(response[0], list):
        segment_id, iot_id = response[0], response[1]
        return {segment_id: str(iot_id)}

    segment_map: dict[int, str] = {}
    for part in response:
        if not isinstance(part, list) or len(part) < 2:
            _LOGGER.warning("Unexpected room mapping entry format: %r", part)
            continue
        segment_id, iot_id = part[0], part[1]
        segment_map[segment_id] = str(iot_id)
    return segment_map



"""Roborock API.

.. include:: ../README.md
"""

from roborock.data import *
from roborock.exceptions import *
from roborock.roborock_typing import *

from . import (
    const,
    data,
    devices,
    exceptions,
    roborock_typing,
    web_api,
)

__all__ = [
    "devices",
    "data",
    "map",
    "web_api",
    "roborock_typing",
    "exceptions",
    "const",
]



"""Create traits for A01 devices.

This module provides the API implementations for A01 protocol devices, which include
Dyad (Wet/Dry Vacuums) and Zeo (Washing Machines).

Using A01 APIs
--------------
A01 devices expose a single API object that handles all device interactions. This API is
available on the device instance (typically via `device.a01_properties`).

The API provides two main methods:
1.  **query_values(protocols)**: Fetches current state for specific data points.
    You must pass a list of protocol enums (e.g. `RoborockDyadDataProtocol` or
    `RoborockZeoProtocol`) to request specific data.
2.  **set_value(protocol, value)**: Sends a command to the device to change a setting
    or perform an action.

Note that these APIs fetch data directly from the device upon request and do not
cache state internally.
"""

import json
from collections.abc import Callable
from datetime import time
from typing import Any

from roborock.data import DyadProductInfo, DyadSndState, HomeDataProduct, RoborockCategory
from roborock.data.dyad.dyad_code_mappings import (
    DyadBrushSpeed,
    DyadCleanMode,
    DyadError,
    DyadSelfCleanLevel,
    DyadSelfCleanMode,
    DyadSuction,
    DyadWarmLevel,
    DyadWaterLevel,
    RoborockDyadStateCode,
)
from roborock.data.zeo.zeo_code_mappings import (
    ZeoDetergentType,
    ZeoDryingMode,
    ZeoError,
    ZeoMode,
    ZeoProgram,
    ZeoRinse,
    ZeoSoftenerType,
    ZeoSpin,
    ZeoState,
    ZeoTemperature,
)
from roborock.devices.rpc.a01_channel import send_decoded_command
from roborock.devices.traits import Trait
from roborock.devices.transport.mqtt_channel import MqttChannel
from roborock.roborock_message import RoborockDyadDataProtocol, RoborockZeoProtocol

__init__ = [
    "DyadApi",
    "ZeoApi",
]


DYAD_PROTOCOL_ENTRIES: dict[RoborockDyadDataProtocol, Callable] = {
    RoborockDyadDataProtocol.STATUS: lambda val: RoborockDyadStateCode(val).name,
    RoborockDyadDataProtocol.SELF_CLEAN_MODE: lambda val: DyadSelfCleanMode(val).name,
    RoborockDyadDataProtocol.SELF_CLEAN_LEVEL: lambda val: DyadSelfCleanLevel(val).name,
    RoborockDyadDataProtocol.WARM_LEVEL: lambda val: DyadWarmLevel(val).name,
    RoborockDyadDataProtocol.CLEAN_MODE: lambda val: DyadCleanMode(val).name,
    RoborockDyadDataProtocol.SUCTION: lambda val: DyadSuction(val).name,
    RoborockDyadDataProtocol.WATER_LEVEL: lambda val: DyadWaterLevel(val).name,
    RoborockDyadDataProtocol.BRUSH_SPEED: lambda val: DyadBrushSpeed(val).name,
    RoborockDyadDataProtocol.POWER: lambda val: int(val),
    RoborockDyadDataProtocol.AUTO_DRY: lambda val: bool(val),
    RoborockDyadDataProtocol.MESH_LEFT: lambda val: int(360000 - val * 60),
    RoborockDyadDataProtocol.BRUSH_LEFT: lambda val: int(360000 - val * 60),
    RoborockDyadDataProtocol.ERROR: lambda val: DyadError(val).name,
    RoborockDyadDataProtocol.VOLUME_SET: lambda val: int(val),
    RoborockDyadDataProtocol.STAND_LOCK_AUTO_RUN: lambda val: bool(val),
    RoborockDyadDataProtocol.AUTO_DRY_MODE: lambda val: bool(val),
    RoborockDyadDataProtocol.SILENT_DRY_DURATION: lambda val: int(val),  # in minutes
    RoborockDyadDataProtocol.SILENT_MODE: lambda val: bool(val),
    RoborockDyadDataProtocol.SILENT_MODE_START_TIME: lambda val: time(
        hour=int(val / 60), minute=val % 60
    ),  # in minutes since 00:00
    RoborockDyadDataProtocol.SILENT_MODE_END_TIME: lambda val: time(
        hour=int(val / 60), minute=val % 60
    ),  # in minutes since 00:00
    RoborockDyadDataProtocol.RECENT_RUN_TIME: lambda val: [
        int(v) for v in val.split(",")
    ],  # minutes of cleaning in past few days.
    RoborockDyadDataProtocol.TOTAL_RUN_TIME: lambda val: int(val),
    RoborockDyadDataProtocol.SND_STATE: lambda val: DyadSndState.from_dict(val),
    RoborockDyadDataProtocol.PRODUCT_INFO: lambda val: DyadProductInfo.from_dict(val),
}

ZEO_PROTOCOL_ENTRIES: dict[RoborockZeoProtocol, Callable] = {
    # read-only
    RoborockZeoProtocol.STATE: lambda val: ZeoState(val).name,
    RoborockZeoProtocol.COUNTDOWN: lambda val: int(val),
    RoborockZeoProtocol.WASHING_LEFT: lambda val: int(val),
    RoborockZeoProtocol.ERROR: lambda val: ZeoError(val).name,
    RoborockZeoProtocol.TIMES_AFTER_CLEAN: lambda val: int(val),
    RoborockZeoProtocol.DETERGENT_EMPTY: lambda val: bool(val),
    RoborockZeoProtocol.SOFTENER_EMPTY: lambda val: bool(val),
    # read-write
    RoborockZeoProtocol.MODE: lambda val: ZeoMode(val).name,
    RoborockZeoProtocol.PROGRAM: lambda val: ZeoProgram(val).name,
    RoborockZeoProtocol.TEMP: lambda val: ZeoTemperature(val).name,
    RoborockZeoProtocol.RINSE_TIMES: lambda val: ZeoRinse(val).name,
    RoborockZeoProtocol.SPIN_LEVEL: lambda val: ZeoSpin(val).name,
    RoborockZeoProtocol.DRYING_MODE: lambda val: ZeoDryingMode(val).name,
    RoborockZeoProtocol.DETERGENT_TYPE: lambda val: ZeoDetergentType(val).name,
    RoborockZeoProtocol.SOFTENER_TYPE: lambda val: ZeoSoftenerType(val).name,
    RoborockZeoProtocol.SOUND_SET: lambda val: bool(val),
}


def convert_dyad_value(protocol_value: RoborockDyadDataProtocol, value: Any) -> Any:
    """Convert a dyad protocol value to its corresponding type."""
    if (converter := DYAD_PROTOCOL_ENTRIES.get(protocol_value)) is not None:
        try:
            return converter(value)
        except (ValueError, TypeError):
            return None
    return None


def convert_zeo_value(protocol_value: RoborockZeoProtocol, value: Any) -> Any:
    """Convert a zeo protocol value to its corresponding type."""
    if (converter := ZEO_PROTOCOL_ENTRIES.get(protocol_value)) is not None:
        try:
            return converter(value)
        except (ValueError, TypeError):
            return None
    return None


class DyadApi(Trait):
    """API for interacting with Dyad devices."""

    def __init__(self, channel: MqttChannel) -> None:
        """Initialize the Dyad API."""
        self._channel = channel

    async def query_values(self, protocols: list[RoborockDyadDataProtocol]) -> dict[RoborockDyadDataProtocol, Any]:
        """Query the device for the values of the given Dyad protocols."""
        response = await send_decoded_command(
            self._channel,
            {RoborockDyadDataProtocol.ID_QUERY: protocols},
            value_encoder=json.dumps,
        )
        return {protocol: convert_dyad_value(protocol, response.get(protocol)) for protocol in protocols}

    async def set_value(self, protocol: RoborockDyadDataProtocol, value: Any) -> dict[RoborockDyadDataProtocol, Any]:
        """Set a value for a specific protocol on the device."""
        params = {protocol: value}
        return await send_decoded_command(self._channel, params)


class ZeoApi(Trait):
    """API for interacting with Zeo devices."""

    name = "zeo"

    def __init__(self, channel: MqttChannel) -> None:
        """Initialize the Zeo API."""
        self._channel = channel

    async def query_values(self, protocols: list[RoborockZeoProtocol]) -> dict[RoborockZeoProtocol, Any]:
        """Query the device for the values of the given protocols."""
        response = await send_decoded_command(
            self._channel,
            {RoborockZeoProtocol.ID_QUERY: protocols},
            value_encoder=json.dumps,
        )
        return {protocol: convert_zeo_value(protocol, response.get(protocol)) for protocol in protocols}

    async def set_value(self, protocol: RoborockZeoProtocol, value: Any) -> dict[RoborockZeoProtocol, Any]:
        """Set a value for a specific protocol on the device."""
        params = {protocol: value}
        return await send_decoded_command(self._channel, params, value_encoder=lambda x: x)


def create(product: HomeDataProduct, mqtt_channel: MqttChannel) -> DyadApi | ZeoApi:
    """Create traits for A01 devices."""
    match product.category:
        case RoborockCategory.WET_DRY_VAC:
            return DyadApi(mqtt_channel)
        case RoborockCategory.WASHING_MACHINE:
            return ZeoApi(mqtt_channel)
        case _:
            raise NotImplementedError(f"Unsupported category {product.category}")



"""
CLI 入口

子命令模式：
  roborock-mcp          → 默认启动 stdio server
  roborock-mcp auth     → 运行交互式认证
  roborock-mcp --help   → 显示帮助
"""

import argparse
import sys

from . import __version__


def main():
    parser = argparse.ArgumentParser(
        prog="roborock-mcp",
        description="Roborock MCP Server — 让 Claude 控制石头扫地机",
        epilog="不带子命令时默认启动 MCP server（stdio 模式）",
    )
    parser.add_argument("--version", "-V", action="version", version=f"%(prog)s {__version__}")
    subparsers = parser.add_subparsers(dest="command")
    subparsers.add_parser("auth", help="首次认证（交互式邮箱验证码）")

    args = parser.parse_args()

    if args.command == "auth":
        from .setup_auth import main as auth_main
        auth_main()
    elif args.command is None:
        # 默认行为：启动 server
        import anyio
        from .server import mcp

        # Windows 的 ProactorEventLoop 不支持 add_reader/add_writer，
        # 而 aiomqtt (paho-mqtt) 依赖这些方法。
        # mcp.run() 内部调用 anyio.run() 但不传 loop_factory，
        # 所以直接调用 anyio.run() 显式指定 SelectorEventLoop。
        # MCP stdio transport 用 anyio.wrap_file()（线程模式），不受影响。
        if sys.platform == "win32":
            import asyncio
            backend_options = {"loop_factory": asyncio.SelectorEventLoop}
        else:
            backend_options = {}

        anyio.run(mcp.run_stdio_async, backend_options=backend_options)
    else:
        # 防御性编程：argparse 通常会拦截未知子命令
        parser.error(f"未知子命令: {args.command}")


if __name__ == "__main__":
    main()



"""Trait for smart wash parameters."""

from roborock.data import SmartWashParams
from roborock.device_features import is_wash_n_fill_dock
from roborock.devices.traits.v1 import common
from roborock.roborock_typing import RoborockCommand


class SmartWashParamsTrait(SmartWashParams, common.V1TraitMixin):
    """Trait for smart wash parameters."""

    command = RoborockCommand.GET_SMART_WASH_PARAMS
    requires_dock_type = is_wash_n_fill_dock



"""Roborock B01 Protocol encoding and decoding."""

import json
import logging
from typing import Any

from roborock.data.b01_q10.b01_q10_code_mappings import B01_Q10_DP
from roborock.exceptions import RoborockException
from roborock.roborock_message import (
    RoborockMessage,
    RoborockMessageProtocol,
)

_LOGGER = logging.getLogger(__name__)

B01_VERSION = b"B01"
ParamsType = list | dict | int | None


def encode_mqtt_payload(command: B01_Q10_DP, params: ParamsType) -> RoborockMessage:
    """Encode payload for B01 Q10 commands over MQTT.

    This does not perform any special encoding for the command parameters and expects
    them to already be in a request specific format.
    """
    dps_data = {
        "dps": {
            # Important: some commands use falsy values so only default to `{}` when params is actually None.
            command.code: params if params is not None else {},
        }
    }
    return RoborockMessage(
        protocol=RoborockMessageProtocol.RPC_REQUEST,
        version=B01_VERSION,
        payload=json.dumps(dps_data).encode("utf-8"),
    )


def _convert_datapoints(datapoints: dict[str, Any], message: RoborockMessage) -> dict[B01_Q10_DP, Any]:
    """Convert the 'dps' dictionary keys from strings to B01_Q10_DP enums."""
    result: dict[B01_Q10_DP, Any] = {}
    for key, value in datapoints.items():
        try:
            code = int(key)
        except ValueError as e:
            raise ValueError(f"dps key is not a valid integer: {e} for {message.payload!r}") from e
        if (dps := B01_Q10_DP.from_code_optional(code)) is not None:
            result[dps] = value
    return result


def decode_rpc_response(message: RoborockMessage) -> dict[B01_Q10_DP, Any]:
    """Decode a B01 Q10 RPC_RESPONSE message.

    This does not perform any special decoding for the response body, but does
    convert the 'dps' keys from strings to B01_Q10_DP enums.
    """
    if not message.payload:
        raise RoborockException("Invalid B01 message format: missing payload")
    try:
        payload = json.loads(message.payload.decode())
    except (json.JSONDecodeError, UnicodeDecodeError) as e:
        raise RoborockException(f"Invalid B01 json payload: {e} for {message.payload!r}") from e

    if (datapoints := payload.get("dps")) is None:
        raise RoborockException(f"Invalid B01 json payload: missing 'dps' for {message.payload!r}")
    if not isinstance(datapoints, dict):
        raise RoborockException(f"Invalid B01 message format: 'dps' should be a dictionary for {message.payload!r}")

    try:
        result = _convert_datapoints(datapoints, message)
    except ValueError as e:
        raise RoborockException(f"Invalid B01 message format: {e}") from e

    # The COMMON response contains nested datapoints need conversion. To simplify
    # response handling at higher levels we flatten these into the main result.
    if B01_Q10_DP.COMMON in result:
        common_result = result.pop(B01_Q10_DP.COMMON)
        if not isinstance(common_result, dict):
            raise RoborockException(f"Invalid dpCommon format: expected dict, got {type(common_result).__name__}")
        try:
            common_dps_result = _convert_datapoints(common_result, message)
        except ValueError as e:
            raise RoborockException(f"Invalid dpCommon format: {e}") from e
        result.update(common_dps_result)

    return result
